from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities import Server, Log
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db
from tests.helpers import TestCaseLockBypass


class Testloglist(TestCaseLockBypass):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        db.create_all()
        set_initial()
        self.node1 = Server('node1', port=8000)
        self.node2 = Server('node2', port=8000)
        self.log = Log(source_server=Server.get_current(), target='/var/log/log1.log', destination_server=self.node1)
        db.session.add_all([self.node1, self.node2, self.log])
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_get(self):
        resp = self.client.get(url_for('api_1_0.loglist'), headers=self.auth.header)
        self.assertListEqual([self.log.to_json()], resp.get_json())

        log = Log(source_server=Server.get_current(), target='/access.log', destination_server=self.node2)
        db.session.add(log)
        db.session.commit()

        # test with filter
        resp = self.client.get(url_for('api_1_0.loglist') + "?filter[target]=/access.log",
                               headers=self.auth.header)
        self.assertListEqual([log.to_json()], resp.get_json())

        # test with filter on a server
        resp = self.client.get(url_for('api_1_0.loglist') + f"?filter[dst_server_id]={self.node1.id}",
                               headers=self.auth.header)
        self.assertListEqual([self.log.to_json()], resp.get_json())

    def test_post(self):
        new_log_json = {"src_server_id": str(self.node1.id),
                        "target": '/var/log',
                        "exclude": 'system.log'}

        resp = self.client.post(url_for('api_1_0.loglist'), headers=self.auth.header,
                                json=new_log_json)
        self.assertEqual(400, resp.status_code)

        new_log_json = {"src_server_id": str(self.node1.id),
                        "target": '/var/log',
                        "exclude": 'system.log',
                        "dst_server_id": str(self.node1.id)}

        resp = self.client.post(url_for('api_1_0.loglist'), headers=self.auth.header,
                                json=new_log_json)
        self.assertEqual(400, resp.status_code)
        self.assertDictEqual({'error': 'source and destination must be different'}, resp.get_json())

        new_log_json = {"src_server_id": str(self.node1.id),
                        "target": '/var/log',
                        "exclude": 'system.log',
                        "dst_server_id": str(self.node2.id)}

        resp = self.client.post(url_for('api_1_0.loglist'), headers=self.auth.header,
                                json=new_log_json)
        self.assertEqual(201, resp.status_code)
        log = Log.query.get(resp.get_json().get('log_id'))
        self.assertEqual(self.node1, log.source_server)
        self.assertEqual('/var/log', log.target)
        self.assertIsNone(log.include)
        self.assertEqual('system.log', log.exclude)
        self.assertEqual(self.node2, log.destination_server)
        self.assertFalse(log.recursive)
        self.assertIsNone(log.dest_folder)


class TestLogResource(TestCaseLockBypass):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        db.create_all()
        set_initial()
        self.node1 = Server('node1', port=8000)
        self.node2 = Server('node2', port=8000)
        self.log = Log(source_server=Server.get_current(), target='/var/log/log1.log',
                       destination_server=self.node1)
        db.session.add_all([self.node1, self.node2, self.log])
        db.session.commit()

    def test_get(self):
        resp = self.client.get(url_for('api_1_0.logresource', log_id=str(self.log.id)), headers=self.auth.header)
        self.assertEqual(200, resp.status_code)

        self.assertEqual(self.log.to_json(), resp.get_json())

        resp = self.client.get(url_for('api_1_0.logresource', log_id='aaaa'), headers=self.auth.header)
        self.assertEqual(404, resp.status_code)

    def test_patch(self):
        patch_log_json = {"src_server_id": str(self.node2.id), }

        resp = self.client.patch(url_for('api_1_0.logresource', log_id=str(self.log.id)), headers=self.auth.header,
                                 json=patch_log_json)
        self.assertEqual(400, resp.status_code)

        patch_log_json = {"dest_folder": '/dest'}

        self.assertIsNone(self.log.dest_folder)
        resp = self.client.patch(url_for('api_1_0.logresource', log_id=str(self.log.id)), headers=self.auth.header,
                                 json=patch_log_json)
        self.assertEqual(204, resp.status_code)
        db.session.refresh(self.log)
        self.assertEqual('/dest', self.log.dest_folder)

        resp = self.client.patch(url_for('api_1_0.logresource', log_id=str(self.log.id)), headers=self.auth.header,
                                 json=patch_log_json)
        self.assertEqual(202, resp.status_code)

    # def test_delete(self):
    #     resp = self.client.delete(url_for('api_1_0.logresource', log_id=str(self.log.id)), headers=self.auth.header)
    #     self.assertEqual(204, resp.status_code)
    #
    #     self.assertEqual(0, Log.query.count())
    #
    #     resp = self.client.delete(url_for('api_1_0.logresource', log_id=str(self.log.id)), headers=self.auth.header)
    #     self.assertEqual(404, resp.status_code)



from unittest import TestCase

from flask_jwt_extended import create_access_token

from dm.domain.entities import Step, ActionTemplate, ActionType, Orchestration
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db, errors


class TestOrchestration(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        db.create_all()
        self.at = ActionTemplate(name='action', version=1, action_type=ActionType.SHELL, code='code to run',
                                 parameters={'param1': 'test'}, expected_stdout='expected output', expected_rc=0,
                                 system_kwargs={})

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_orch_creation(self):
        o = Orchestration(id="aaaaaaaa-1234-5678-1234-56781234aaa1",
                          name='Test Orchestration',
                          version=1,
                          description='description',
                          params={'param1': 2})

        db.session.add(o)
        db.session.commit()
        del o
        o = Orchestration.query.get("aaaaaaaa-1234-5678-1234-56781234aaa1")

    def test_parameters(self):
        o = Orchestration(id=1,
                          name='Test Orchestration',
                          version=1,
                          description='description')

        s1 = o.add_step(undo=False, action_template=self.at, parents=[], children=[], stop_on_error=False, id=1)

        at = ActionTemplate(name='action', version=1, action_type=ActionType.SHELL, code='code to run',
                            parameters={}, expected_stdout='expected output', expected_rc=0,
                            system_kwargs={})

        self.assertDictEqual(dict(s1.parameters), {'param1': 'test'})

        s1.action_template = at

        self.assertDictEqual(dict(s1.parameters), {})

    def test_set_dependencies(self):
        s1 = Step(orchestration=None, undo=False, stop_on_error=False, action_template=self.at,
                  id='11111111-2222-3333-4444-555555550001')
        s2 = Step(orchestration=None, undo=False, stop_on_error=False, action_template=self.at,
                  id='11111111-2222-3333-4444-555555550002')
        s3 = Step(orchestration=None, undo=False, stop_on_error=False, action_template=self.at,
                  id='11111111-2222-3333-4444-555555550003')

        o = Orchestration('test', 1, id='11111111-2222-3333-4444-555555550004',
                          description='description test', steps=[s1, s2, s3],
                          dependencies={s1.id: [s2.id], s2.id: [s3.id], s3.id: []})

        self.assertDictEqual(o.dependencies, {s1: [s2], s2: [s3], s3: []})

        self.assertListEqual(o.steps, [s1, s2, s3])

        with self.assertRaises(ValueError):
            o.set_dependencies({s1: [s2], s2: [s3], s3: [4]})

        o = Orchestration('test', 1, id='11111111-2222-3333-4444-555555550004',
                          description='description test', steps=[s1, s2, s3])

        self.assertListEqual([s1, s2, s3], o.steps)
        o.set_dependencies([(s1, s2), (s2, s3)])
        self.assertDictEqual(o.dependencies, {s1: [s2], s2: [s3], s3: []})

    # noinspection PyTypeChecker
    def test_attributes_and_methods(self):
        o = Orchestration(id=1,
                          name='Test Orchestration',
                          version=1,
                          description='description')

        s1 = o.add_step(undo=False, action_template=self.at, parents=[], children=[], stop_on_error=False,
                        id=1)
        s2 = o.add_step(undo=True, action_template=self.at, parents=[s1], children=[], stop_on_error=False,
                        id=2)

        with self.assertRaises(errors.ParentUndoError):
            o.add_step(undo=False, action_template=self.at, parents=[s2], children=[], stop_on_error=False)
        self.assertListEqual(o.steps, [s1, s2])

        o2 = Orchestration('dto', 2, id=2)
        s21 = o2.add_step(undo=False, action_template=self.at, parents=[], children=[], stop_on_error=False,
                          id=21)

        with self.assertRaises(ValueError):
            o.add_parents(s21, [s1])

        o.delete_step(s2)

        self.assertListEqual(o.steps, [s1])
        self.assertListEqual(o.children[s1], [])

        s2 = o.add_step(undo=False, action_template=self.at, parents=[s1], children=[], stop_on_error=False, id=2)
        s3 = o.add_step(undo=False, action_template=self.at, parents=[s2], children=[], stop_on_error=False, id=3)
        s4 = o.add_step(undo=False, action_template=self.at, parents=[s2], children=[], stop_on_error=False, id=4)
        s5 = o.add_step(undo=False, action_template=self.at, parents=[s4], children=[], stop_on_error=False, id=5)
        s6 = o.add_step(undo=False, action_template=self.at, parents=[s4], children=[], stop_on_error=False, id=6)
        s7 = o.add_step(undo=False, action_template=self.at, parents=[s1], children=[s2], stop_on_error=False,
                        id=7)

        self.assertListEqual(o.steps, [s1, s2, s3, s4, s5, s6, s7])

        self.assertDictEqual(o.children,
                             {s1: [s2, s7], s2: [s3, s4], s3: [], s4: [s5, s6], s5: [], s6: [], s7: [s2]})
        self.assertDictEqual(o.parents,
                             {s1: [], s2: [s1, s7], s3: [s2], s4: [s2], s5: [s4], s6: [s4], s7: [s1]})

        with self.assertRaises(errors.ChildDoError):
            o.add_step(undo=True, action_template=self.at, parents=[s1], children=[s2], stop_on_error=False)

        with self.assertRaises(errors.CycleError):
            o.add_step(undo=False, action_template=self.at, parents=[s6], children=[s1], stop_on_error=False)

        # Check parent functions
        o.add_parents(s6, [s2])
        self.assertListEqual(o.children[s2], [s3, s4, s6])

        o.set_parents(s6, [s3, s4])
        self.assertListEqual(o.parents[s6], [s3, s4])

        o.delete_parents(s6, [s3, s2])
        self.assertListEqual(o.parents[s6], [s4])

        # Check children functions
        o.add_children(s3, [s6])
        self.assertListEqual(o.parents[s6], [s4, s3])

        o.delete_children(s4, [s5, s6])
        self.assertListEqual(o.parents[s6], [s3])
        self.assertListEqual(o.parents[s5], [])

        o.set_children(s4, [s5, s6]).set_children(s3, [])
        self.assertListEqual([s4], o.parents[s6])
        self.assertListEqual([s4], o.parents[s5])
        self.assertListEqual([], o.children[s3])

        # properties and default values
        s = o.add_step(undo=False, action_template=self.at)
        self.assertEqual(True, s.stop_on_error)
        self.assertEqual('code to run', s.code)
        self.assertEqual('expected output', s.expected_stdout)
        s.expected_output = 'changed'
        self.assertEqual('changed', s.expected_output)
        s.expected_output = ''
        self.assertEqual('', s.expected_output)
        s.expected_output = None
        self.assertEqual('expected output', s.expected_stdout)
        self.assertEqual(0, s.expected_rc)
        s.expected_rc = 2
        self.assertEqual(2, s.expected_rc)
        s.expected_rc = 0
        self.assertEqual(0, s.expected_rc)
        s.expected_rc = None
        self.assertEqual(0, s.expected_rc)

    def test_eq_imp(self):
        o1 = Orchestration('dto', 1)

        s11 = o1.add_step(False, self.at)
        s12 = o1.add_step(False, self.at, parents=[s11])

        o2 = Orchestration('dto', 2)
        s21 = o2.add_step(False, self.at, )
        s22 = o2.add_step(False, self.at, parents=[s21])

        self.assertTrue(o1.eq_imp(o2))
        self.assertTrue(o2.eq_imp(o1))

        s23 = o2.add_step(True, self.at, parents=[s22])

        self.assertFalse(o1.eq_imp(o2))
        self.assertFalse(o2.eq_imp(o1))

        s13 = o1.add_step(True, self.at, parents=[s12])

        self.assertTrue(o1.eq_imp(o2))

        s13.step_parameters['server'] = 'localhost'
        self.assertFalse(s13.eq_imp(s23))
        self.assertFalse(o1.eq_imp(o2))

    def test_init_on_load(self):
        o = Orchestration(id='aaaaaaaa-1234-5678-1234-aaaaaaaa0001',
                          name='Test Orchestration',
                          version=1,
                          description='description')

        s1 = o.add_step(undo=False, action_template=self.at, parents=[], children=[], stop_on_error=False,
                        id='bbbbbbbb-1234-5678-1234-bbbbbbbb0001')

        db.session.add(o)
        db.session.commit()
        del o

        o = Orchestration.query.get('aaaaaaaa-1234-5678-1234-aaaaaaaa0001')

        self.assertEqual({s1: []}, o._graph.succ)


        s2 = o.add_step(undo=False, action_template=self.at, parents=[], children=[], stop_on_error=False,
                        id='bbbbbbbb-1234-5678-1234-bbbbbbbb0002')
        del o

        o = Orchestration.query.get('aaaaaaaa-1234-5678-1234-aaaaaaaa0001')

        self.assertEqual({s1: [], s2: []}, o._graph.succ)



from setuptools import setup, find_packages

import dm

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='dimensigon',
    version=dm.__version__,
    packages=find_packages(),
    url='https://github.com/dimensigon/dimensigon',
    license=dm.__license__,
    author=dm.__author__,
    author_email=dm.__email__,
    description="Dimensigon (DM) is a Distributed Management and Automation RESTful framework to make easier heterogeneous/polyglot orchestrations in a Hybrid Multi-Cloud environment.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "beautifulsoup4",
        "PyYAML",
        "six",
        "flask",
        "flask-JWT-Extended",
        "flask-Migrate",
        "flask-RESTFul",
        "flask-SQLAlchemy",
        "requests",
        "rsa",
        "coverage",
        "jsonschema",
        "aiohttp",
        "dataclasses",
        "cryptography",
        "psutil",
        "python-dotenv",
        "netifaces",
        "apscheduler",
        "passlib",
        "jinja2",
        "SQLAlchemy",
        "Werkzeug",
        "jinja2schema",
        "RestrictedPython",
        "setuptools",
        "click",
        "pygments"
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX",
    ],
    python_requires='>=3.6',
    entry_points = {'console_scripts': ['dshell=dm.dshell.batch.dshell:main']},
)



import hashlib
import os
import re

from aioresponses import aioresponses, CallbackResult
from flask_jwt_extended import create_access_token
from pyfakefs.fake_filesystem_unittest import TestCase

from dm.domain.entities import Server, Route, Dimension, Transfer, TransferStatus
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.utils.asyncio import run
from dm.utils.helpers import generate_dimension, md5
from dm.web import create_app, db
from dm.web.async_functions import async_send_file
from tests.helpers import set_callbacks


class TestAsyncSendFile(TestCase):
    def setUp(self) -> None:
        # create the app with common test config
        self.app1 = create_app('test')
        self.app1.config['SERVER_NAME'] = 'node1'
        # self.app1.config['SECURIZER'] = True
        self.client1 = self.app1.test_client()
        self.app2 = create_app('test')
        self.app2.config['SERVER_NAME'] = 'node2'
        # self.app2.config['SECURIZER'] = True
        self.client2 = self.app2.test_client()

        with self.app1.app_context():
            db.create_all()
            set_initial()
            s = Server.get_current()
            s.gates = []
            s.add_new_gate('node1', 5000)
            dim = generate_dimension('dimension')
            dim.current = True
            db.session.add(dim)
            db.session.commit()
            self.s1_json = Server.get_current().to_json()
            self.dim_json = dim.to_json()
            self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))

        with self.app2.app_context():
            db.create_all()
            set_initial()
            s = Server.get_current()
            s.gates = []
            s.add_new_gate('node2', 5000)
            db.session.commit()
            self.s2_json = Server.get_current().to_json()
            s = Server.from_json(self.s1_json)
            s.add_new_gate('node1', 5000)
            Route(s, cost=0)
            db.session.add(s)
            dim = Dimension.from_json(self.dim_json)
            dim.current = True
            db.session.add(dim)
            db.session.commit()

        self.context = self.app1.app_context()
        self.context.push()
        s = Server.from_json(self.s2_json)
        s.add_new_gate('node2', 5000)
        Route(s, cost=0)
        db.session.add(s)

        self.remote = s

        self.source_path = '/software'
        self.filename = 'filename.zip'
        self.content = b'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
        self.size = len(self.content)
        self.checksum = hashlib.md5(self.content).hexdigest()
        self.dest_path = '/dest_repo'

        db.session.commit()

        self.setUpPyfakefs()
        self.fs.create_dir(self.source_path)
        self.fs.create_dir(self.dest_path)
        self.fs.create_file(os.path.join(self.source_path, self.filename), contents=self.content)

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.context.pop()

        with self.app2.app_context():
            db.session.remove()
            db.drop_all()

    @aioresponses()
    def test_async_send_file(self, m):
        set_callbacks([(r'(node1|127\.0\.0\.1)', self.client1), ('node2', self.client2)], m)

        with self.app2.app_context():
            transfer = Transfer(software=self.filename,
                                size=self.size,
                                checksum=self.checksum,
                                dest_path=self.dest_path,
                                num_chunks=16,
                                status=TransferStatus.WAITING_CHUNKS)
            db.session.add(transfer)
            db.session.commit()
            transfer_id = transfer.id

        self.assertFalse(os.path.exists(os.path.join(self.dest_path, self.filename)))

        run(async_send_file(dest_server=self.remote, transfer_id=transfer_id,
                            file=os.path.join(self.source_path, self.filename), chunk_size=4,
                            auth=self.auth))

        self.assertTrue(os.path.exists(os.path.join(self.dest_path, self.filename)))
        self.assertEqual(self.size, os.path.getsize(os.path.join(self.dest_path, self.filename)))
        self.assertEqual(self.checksum, md5(os.path.join(self.dest_path, self.filename)))

    @aioresponses()
    def test_async_send_file_one_chunk(self, m):
        set_callbacks([(r'(node1|127\.0\.0\.1)', self.client1), ('node2', self.client2)], m)

        with self.app2.app_context():
            transfer = Transfer(software=self.filename,
                                size=self.size,
                                checksum=self.checksum,
                                dest_path=self.dest_path,
                                num_chunks=1,
                                status=TransferStatus.WAITING_CHUNKS)
            db.session.add(transfer)
            db.session.commit()
            transfer_id = transfer.id

        self.assertFalse(os.path.exists(os.path.join(self.dest_path, self.filename)))

        run(async_send_file(dest_server=self.remote, transfer_id=transfer_id,
                            file=os.path.join(self.source_path, self.filename), chunk_size=80,
                            auth=self.auth))

        self.assertTrue(os.path.exists(os.path.join(self.dest_path, self.filename)))
        self.assertEqual(self.size, os.path.getsize(os.path.join(self.dest_path, self.filename)))
        self.assertEqual(self.checksum, md5(os.path.join(self.dest_path, self.filename)))

    @aioresponses()
    def test_async_send_retry(self, m):
        def post_callback_client(url, **kwargs):
            kwargs.pop('allow_redirects')

            r = self.client2.post(url.path, json=kwargs['json'], headers=kwargs['headers'])

            return CallbackResult('POST', status=r.status_code, body=r.data, content_type=r.content_type,
                                  headers=r.headers)

        m.post(re.compile(Server.query.filter_by(name='node2').one().url() + '.*'), callback=post_callback_client)
        m.post(re.compile(Server.query.filter_by(name='node2').one().url() + '.*'), callback=post_callback_client)
        m.post(re.compile(Server.query.filter_by(name='node2').one().url() + '.*'), exception=ConnectionError)
        m.post(re.compile(Server.query.filter_by(name='node2').one().url() + '.*'), exception=ConnectionError)
        m.post(re.compile(Server.query.filter_by(name='node2').one().url() + '.*'), callback=post_callback_client)
        m.post(re.compile(Server.query.filter_by(name='node2').one().url() + '.*'), callback=post_callback_client)
        m.post(re.compile(Server.query.filter_by(name='node2').one().url() + '.*'), exception=ConnectionError)
        m.post(re.compile(Server.query.filter_by(name='node2').one().url() + '.*'), callback=post_callback_client)

        def patch_callback_client(url, **kwargs):
            kwargs.pop('allow_redirects')

            r = self.client2.patch(url.path, json=kwargs['json'], headers=kwargs['headers'])

            return CallbackResult('PATCH', status=r.status_code, body=r.data, content_type=r.content_type,
                                  headers=r.headers)

        m.patch(re.compile(Server.query.filter_by(name='node2').one().url() + '.*'), callback=patch_callback_client,
                repeat=True)

        with self.app2.app_context():
            transfer = Transfer(software=self.filename,
                                size=self.size,
                                checksum=self.checksum,
                                dest_path=self.dest_path,
                                num_chunks=5,
                                status=TransferStatus.WAITING_CHUNKS)
            db.session.add(transfer)
            db.session.commit()
            transfer_id = transfer.id

        self.assertFalse(os.path.exists(os.path.join(self.dest_path, self.filename)))

        run(async_send_file(dest_server=self.remote, transfer_id=transfer_id,
                            file=os.path.join(self.source_path, self.filename), chunk_size=14,
                            auth=self.auth, retries=3))

        self.assertTrue(os.path.exists(os.path.join(self.dest_path, self.filename)))
        self.assertEqual(self.size, os.path.getsize(os.path.join(self.dest_path, self.filename)))
        self.assertEqual(self.checksum, md5(os.path.join(self.dest_path, self.filename)))




import copy
import os
import typing as t

from dm.domain.entities import Server
from dm.domain.entities.base import DistributedEntityMixin, UUIDistributedEntityMixin
from dm.utils.typos import UUID
from dm.web import db


class SoftwareServerAssociation(db.Model, DistributedEntityMixin):
    __tablename__ = "D_software_server"
    order = 20

    software_id = db.Column(UUID, db.ForeignKey("D_software.id"), primary_key=True, nullable=False)
    server_id = db.Column(UUID, db.ForeignKey("D_server.id"), primary_key=True, nullable=False)
    path = db.Column(db.Text, nullable=False)

    software = db.relationship("Software", back_populates="ssas", uselist=False)
    server = db.relationship("Server", backref="software_list", uselist=False)

    def to_json(self):
        data = super().to_json()
        data.update({'software_id': str(self.software.id), 'server_id': str(self.server.id), 'path': self.path})
        return data

    @classmethod
    def from_json(cls, kwargs):
        kwargs = copy.deepcopy(kwargs)
        kwargs['software'] = Software.query.get(kwargs.get('software_id'))
        kwargs['server'] = Server.query.get(kwargs.get('server_id'))
        super().from_json(kwargs)
        try:
            o = cls.query.get((kwargs['software_id'], kwargs['server_id']))
        except RuntimeError as e:
            o = None
        if o:
            for k, v in kwargs.items():
                if getattr(o, k) != v:
                    setattr(o, k, v)
            return o
        else:
            return cls(**kwargs)

    @property
    def file(self):
        return os.path.join(self.path or '', self.software.filename or '')


class Software(db.Model, UUIDistributedEntityMixin):
    __tablename__ = "D_software"
    order = 10

    name = db.Column(db.String(80), nullable=False)
    version = db.Column(db.String(40), nullable=False)
    family = db.Column(db.String(50))
    filename = db.Column(db.String(256))
    size = db.Column(db.Integer)
    checksum = db.Column(db.Text())

    ssas: t.List[SoftwareServerAssociation]= db.relationship("SoftwareServerAssociation", back_populates="software")

    __table_args__ = (
        db.UniqueConstraint('name', 'version', name='D_software_u01'),)

    def __init__(self, name, version, filename, family=None, size=None, checksum=None, **kwargs):
        UUIDistributedEntityMixin.__init__(self, **kwargs)
        self.name = name
        self.version = version
        self.family = family
        self.filename = filename
        self.size = size
        self.checksum = checksum

    def to_json(self, servers=False):
        data = super().to_json()
        data.update({'name': self.name, 'version': self.version, 'family': self.family,
                     'filename': self.filename, 'size': self.size, 'checksum': self.checksum})
        if servers:
            servers = []
            for ssa in self.ssas:
                server_data = dict(server=ssa.server.name, server_id=str(ssa.server.id), path=ssa.path)
                servers.append(server_data)
            data.update(servers=servers)
        return data

    @classmethod
    def from_json(cls, kwargs):
        kwargs = copy.deepcopy(kwargs)
        return super().from_json(kwargs)

    def __str__(self):
        return f"{self.name}.{self.version}"



from unittest import TestCase
from unittest.mock import patch

from dm.domain.entities import Server, Route
from dm.web import create_app, db, errors


class TestServer(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        # set_initial()
        db.create_all()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def set_servers_and_routes(self):
        self.n1 = Server(name='n1', dns_or_ip='1.1.1.1')
        self.n2 = Server(name='n2', dns_or_ip='n2_dns')
        self.n3 = Server(name='n3', port=8000)
        self.r1 = Server(name='r1', dns_or_ip='3.3.3.3')
        self.r2 = Server(name='r2', dns_or_ip='r2_dns')
        db.session.add_all([self.n1, self.n2, self.r1, self.r2])
        db.session.commit()

        Route(destination=self.n1, cost=0)
        Route(destination=self.n2, cost=0)
        Route(destination=self.n3, cost=0)
        Route(destination=self.r1, proxy_server=self.n1, cost=1)
        Route(destination=self.r2, proxy_server=self.n2, cost=1)

        db.session.commit()

    @patch('dm.domain.entities.route.check_host')
    @patch('dm.domain.entities.server.url_for')
    def test_url(self, mock_url, mock_check_host):
        self.set_servers_and_routes()

        mock_check_host.return_value = True
        self.assertEqual('http://1.1.1.1:8000', self.n1.url())
        self.assertEqual('http://n2_dns:8000', self.n2.url())
        self.assertEqual('http://n3:8000', self.n3.url())
        self.assertEqual('http://1.1.1.1:8000', self.r1.url())
        self.assertEqual('http://n2_dns:8000', self.r2.url())

        mock_url.return_value = '/'

        self.assertEqual('http://1.1.1.1:8000/', self.n1.url('api'))

        mock_url.assert_called_once_with('api')

        me = Server(name='me', gates=[('127.0.0.1', 5), ('192.168.1.2', 2)], me=True)
        self.assertEqual(f'http://127.0.0.1:5/', me.url('api'))

        self.app.config['PREFERRED_URL_SCHEME'] = 'https'
        me = Server(name='me', gates=[('192.168.1.2', 2)], me=True)
        self.assertEqual(f'https://192.168.1.2:2/', me.url('api'))

        s = Server('test', port=8000)
        with self.assertRaises(errors.UnreachableDestination):
            s.url()

    def test_get_neighbours(self):
        n1 = Server('n1', port=8000)
        n2 = Server('n2', port=8000)
        n3 = Server('n3', port=8000)
        r1 = Server('r1', port=8000)
        Route(destination=n1, cost=0)
        Route(destination=n2, gate=n2.gates[0])
        Route(destination=r1, proxy_server=n1, cost=1)

        me = Server('me', port=8000, me=True)
        db.session.add_all([n1, n2, n3, r1, me])

        self.assertListEqual([n1, n2], me.get_neighbours())

    def test_get_neighbours_no_route(self):
        n1 = Server('n1', port=8000)
        me = Server('me', port=8000, me=True)

        db.session.add_all([n1, me])

        self.assertListEqual([], me.get_neighbours())

    def test_get_not_neighbours(self):
        n1 = Server('n1', port=8000, id='22cd859d-ee91-4079-a112-000000000001')
        n2 = Server('n2', port=8000, id='22cd859d-ee91-4079-a112-000000000002')
        n3 = Server('n3', port=8000, id='22cd859d-ee91-4079-a112-000000000003')
        n3.route = None
        r1 = Server('r1', port=8000, id='22cd859d-ee91-4079-a112-000000000011')
        Route(destination=n1, cost=0)
        Route(destination=n2, gate=n2.gates[0])
        Route(destination=r1, proxy_server=n1, cost=1)

        me = Server('me', port=8000, me=True)
        db.session.add_all([n1, n2, n3, r1, me])

        self.assertListEqual([n3, r1], me.get_not_neighbours())

    @patch('dm.domain.entities.base.uuid.uuid4')
    def test_from_to_json_with_gate(self, mock_uuid):
        mock_uuid.return_value = '22cd859d-ee91-4079-a112-000000000002'
        s = Server('server2', gates=[('dns', 6000)], id='22cd859d-ee91-4079-a112-000000000001')
        self.assertDictEqual(
            {'id': '22cd859d-ee91-4079-a112-000000000001', 'name': 'server2', 'granules': [],
             'gates': [{'id': '22cd859d-ee91-4079-a112-000000000002', 'ip': None, 'dns': 'dns', 'port': 6000,
                        'hidden': False}]},
            s.to_json(
                add_gates=True))
        db.session.add(s)
        db.session.commit()

        s_json = s.to_json(add_gates=True)
        smashed = Server.from_json(s_json)

        self.assertIs(s, smashed)
        self.assertEqual(s.id, smashed.id)
        self.assertEqual(s.name, smashed.name)
        self.assertEqual(s.granules, smashed.granules)
        self.assertEqual(s.last_modified_at, smashed.last_modified_at)
        self.assertListEqual(s.gates, smashed.gates)

        # from new Server
        db.session.remove()
        db.drop_all()
        db.create_all()

        smashed = Server.from_json(s_json)

        self.assertEqual(s.id, smashed.id)
        self.assertEqual(s.name, smashed.name)
        self.assertEqual(s.granules, smashed.granules)
        self.assertEqual(s.last_modified_at, smashed.last_modified_at)
        self.assertEqual(1, len(smashed.gates))
        self.assertEqual(s.gates[0].id, smashed.gates[0].id)
        self.assertEqual(s.gates[0].ip, smashed.gates[0].ip)
        self.assertEqual(s.gates[0].dns, smashed.gates[0].dns)
        self.assertEqual(s.gates[0].port, smashed.gates[0].port)



from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities import Server, User
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db
from tests.helpers import TestCaseLockBypass


class TestLogResourceList(TestCaseLockBypass):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()

        db.create_all()
        s = Server('server', port=8000, me=True)
        db.session.add(s)

        self.user = User(user='user', active=False)
        db.session.add(self.user)

        db.session.commit()
        self.auth = HTTPBearerAuth(create_access_token(self.user.id))

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_get(self):
        resp = self.client.get(url_for('api_1_0.userlist'), headers=self.auth.header)
        self.assertListEqual([self.user.to_json()], resp.get_json())

        root = User(user='root')
        db.session.add(root)
        db.session.commit()

        # test with filter
        resp = self.client.get(url_for('api_1_0.userlist') + "?filter[active]=True",
                               headers=self.auth.header)
        self.assertListEqual([root.to_json()], resp.get_json())

        # test with filter on a server
        resp = self.client.get(url_for('api_1_0.userlist') + f"?filter[user]=root",
                               headers=self.auth.header)
        db.session.refresh(root)
        self.assertListEqual([root.to_json()], resp.get_json())

    def test_post(self):
        new_user_json = {"user": 'root',
                         "password": "1234",
                         "email": 'root@dimensigon.com',
                         "name": "Root"
                         }

        resp = self.client.post(url_for('api_1_0.userlist'), headers=self.auth.header,
                                json=new_user_json)
        self.assertEqual(400, resp.status_code)

        new_user_json = {"user": 'root',
                         "password": "1234",
                         "email": 'root@dimensigon.com',
                         }

        resp = self.client.post(url_for('api_1_0.userlist'), headers=self.auth.header,
                                json=new_user_json)
        self.assertEqual(201, resp.status_code)
        user = User.query.get(resp.get_json().get('user_id'))
        self.assertEqual('root', user.user)
        self.assertEqual("root@dimensigon.com", user.email)
        self.assertIsNotNone(user._password)
        self.assertNotEqual('1234', user._password)
        self.assertTrue(user.active)


class TestUserResource(TestCaseLockBypass):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()
        s = Server('server', port=8000, me=True)
        db.session.add(s)
        self.user = User(user='user', active=False)
        db.session.add(self.user)
        db.session.commit()
        self.auth = HTTPBearerAuth(create_access_token(self.user.id))

    def test_get(self):
        resp = self.client.get(url_for('api_1_0.userresource', user_id=str(self.user.id)), headers=self.auth.header)
        self.assertEqual(200, resp.status_code)

        self.assertEqual(self.user.to_json(), resp.get_json())

        resp = self.client.get(url_for('api_1_0.userresource', user_id='aaaa'), headers=self.auth.header)
        self.assertEqual(404, resp.status_code)

    def test_patch(self):
        patch_user_json = {"name": "Kevin"}

        resp = self.client.patch(url_for('api_1_0.userresource', user_id=str(self.user.id)), headers=self.auth.header,
                                 json=patch_user_json)
        self.assertEqual(400, resp.status_code)
        db.session.refresh(self.user)

        patch_user_json = {"email": "root@dimensigon.com"}

        self.assertIsNone(self.user.email)
        resp = self.client.patch(url_for('api_1_0.userresource', user_id=str(self.user.id)), headers=self.auth.header,
                                 json=patch_user_json)
        self.assertEqual(204, resp.status_code)
        db.session.refresh(self.user)
        self.assertEqual('root@dimensigon.com', self.user.email)

        resp = self.client.patch(url_for('api_1_0.userresource', user_id=str(self.user.id)), headers=self.auth.header,
                                 json=patch_user_json)
        self.assertEqual(202, resp.status_code)

    # def test_delete(self):
    #     resp = self.client.delete(url_for('api_1_0.userresource', user_id=str(self.user.id)), headers=self.auth.header)
    #     self.assertEqual(204, resp.status_code)
    #
    #     self.assertEqual(0, User.query.count())
    #
    #     resp = self.client.delete(url_for('api_1_0.userresource', user_id=str(self.user.id)), headers=self.auth.header)
    #     self.assertEqual(404, resp.status_code)



from unittest import TestCase

from flask_jwt_extended import create_access_token

from dm import defaults
from dm.domain.entities import Server, Software, SoftwareServerAssociation
from dm.domain.entities.bootstrap import set_initial
from dm.web import create_app, db


class TestApi(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        db.create_all()
        set_initial()

        self.soft = Software(id='aaaaaaaa-1234-5678-1234-56781234aaa1', name='test', version='1',
                             filename='file')
        self.soft2 = Software(id='aaaaaaaa-1234-5678-1234-56781234aaa2', name='test', version='2',
                              filename='file')
        self.ssa = SoftwareServerAssociation(software=self.soft, server=Server.get_current(), path='/root')
        db.session.add_all([self.soft, self.soft2, self.ssa])
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_to_from_json(self):
        ssa_json = self.ssa.to_json()
        ssa_json['path'] = '/new_root'

        smashed = SoftwareServerAssociation.from_json(ssa_json)

        self.assertEqual("/new_root", smashed.path)
        self.assertIsNotNone(smashed.path)
        self.assertEqual(Server.get_current(), smashed.server)
        self.assertIsNotNone(smashed.server)
        self.assertEqual(self.soft, smashed.software)
        self.assertIsNotNone(smashed.software)
        self.assertEqual(self.ssa.last_modified_at, smashed.last_modified_at)
        self.assertIsNotNone(smashed.last_modified_at)

        db.session.commit()

        del smashed

        ssa = SoftwareServerAssociation.query.get(('aaaaaaaa-1234-5678-1234-56781234aaa1', Server.get_current().id))
        self.assertEqual("/new_root", ssa.path)

    def test_from_json_new(self):
        ssa_json = dict(software_id='aaaaaaaa-1234-5678-1234-56781234aaa2',
                        server_id=str(Server.get_current().id),
                        path='/root',
                        last_modified_at=defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT))

        smashed = SoftwareServerAssociation.from_json(ssa_json)

        self.assertEqual(defaults.INITIAL_DATEMARK, smashed.last_modified_at)
        self.assertIsNotNone(smashed.last_modified_at)
        self.assertEqual("/root", smashed.path)
        self.assertIsNotNone(smashed.path)
        self.assertEqual(Server.get_current(), smashed.server)
        self.assertIsNotNone(smashed.server)
        self.assertEqual(self.soft2, smashed.software)
        self.assertIsNotNone(smashed.software)



import itertools
import typing as t

from sqlalchemy import orm

from dm.domain.entities.base import UUIDistributedEntityMixin
from dm.utils.dag import DAG
from dm.web import db, errors
from .step import Step
from ...utils.helpers import get_now
from ...utils.typos import UtcDateTime

if t.TYPE_CHECKING:
    from dm.domain.entities import ActionTemplate, ActionType

Tdependencies = t.Union[t.Dict[Step, t.Iterable[Step]], t.Iterable[t.Tuple[Step, Step]]]


class Orchestration(db.Model, UUIDistributedEntityMixin):
    __tablename__ = 'D_orchestration'
    order = 20

    name = db.Column(db.String(80), nullable=False)
    version = db.Column(db.Integer, nullable=False)
    description = db.Column(db.Text)
    stop_on_error = db.Column(db.Boolean)
    stop_undo_on_error = db.Column(db.Boolean)
    undo_on_error = db.Column(db.Boolean)
    parameters = db.Column("parameters", db.JSON)
    created_at = db.Column(UtcDateTime(timezone=True), default=get_now)

    steps = db.relationship("Step", primaryjoin="Step.orchestration_id==Orchestration.id",
                            back_populates="orchestration")

    __table_args__ = (db.UniqueConstraint('name', 'version', name='D_orchestration_uq01'),)

    def __init__(self, name: str, version: int, description: t.Optional[str] = None, steps: t.List[Step] = None,
                 stop_on_error: bool = True, stop_undo_on_error: bool = True, undo_on_error: bool = True,
                 parameters=None,
                 dependencies: Tdependencies = None, created_at=None, **kwargs):

        UUIDistributedEntityMixin.__init__(self, **kwargs)

        self.name = name
        self.version = version
        self.description = description
        self.steps = steps or []
        assert isinstance(stop_on_error, bool)
        self.stop_on_error = stop_on_error
        assert isinstance(stop_undo_on_error, bool)
        self.stop_undo_on_error = stop_undo_on_error
        assert isinstance(undo_on_error, bool)
        self.undo_on_error = undo_on_error
        self.parameters = parameters or {}
        self.created_at = created_at or get_now()

        if dependencies:
            self.set_dependencies(dependencies)
        else:
            self._graph = DAG()

    @orm.reconstructor
    def init_on_load(self):
        self.parameters = self.parameters or {}
        self._graph = DAG()
        for step in self.steps:
            if step.parents:
                for p in step.parents:
                    self._graph.add_edge(p, step)
            else:
                self._graph.add_node(step)

    def set_dependencies(self, dependencies: Tdependencies):
        edges = []
        find = lambda id_: next((step for step in self.steps if step.id == id_))
        if isinstance(dependencies, t.Dict):
            for k, v in dependencies.items():
                try:
                    step_from = find(k)
                except StopIteration:
                    raise ValueError(f'id step {k} not found in steps list')
                for child_id in v:
                    try:
                        step_to = find(child_id)
                    except StopIteration:
                        raise ValueError(f'id step {child_id} not found in steps list')
                    edges.append((step_from, step_to))
        elif isinstance(dependencies, t.Iterable):
            edges = dependencies
        else:
            raise ValueError(f'dependencies must be a dict like object or an iterable of tuples. '
                             f'See the docs for more information')
        self._graph = DAG(edges)

    @property
    def parents(self) -> t.Dict[Step, t.List[Step]]:
        return self._graph.pred

    @property
    def children(self) -> t.Dict[Step, t.List[Step]]:
        return self._graph.succ

    @property
    def dependencies(self) -> t.Dict[Step, t.List[Step]]:
        return self.children

    @property
    def root(self) -> t.List[Step]:
        return self._graph.root

    @property
    def target(self) -> t.Set[str]:
        target = set()
        for step in self.steps:
            if step.undo is False:
                target.update(step.target)
        return target

    @property
    def user_parameters(self) -> t.Set['str']:
        params = set()
        for s in self.steps:
            params = params.union(s.user_parameters)
        return params - set(self.parameters.keys()) - set(itertools.chain(*[list(s.parameters.keys()) for s in self.steps]))

    def _step_exists(self, step: t.Union[t.List[Step], Step]):
        """Checks if all the steps belong to the current orchestration

        Parameters
        ----------
        step: list[Step]|Step
            Step or list of steps to be evaluated

        Returns
        -------
        None

        Raises
        ------
        ValueError: if any step passed as argument is not in the orchestration
        """
        if not isinstance(step, list):
            steps = [step]
        else:
            steps = step

        for step in steps:
            if not (step in self._graph.nodes and step.orchestration is self):
                raise ValueError(f'{step} is NOT from this orchestration')

    def _check_dependencies(self, step, parents=None, children=None):
        """
        Checks if the dependencies that are going to be added accomplish the business rules. These rules are:
            1. a 'do' Step cannot be preceded for an 'undo' Step
            2. cannot be cycles inside the orchestration

        Parameters
        ----------
        step: Step
            step to be evaluated
        parents: list[Step]
            parent steps to be added
        children: list[Step]
            children steps to be added

        Raises
        ------
        ValueError
            if the rule 1 is not passed
        CycleError
            if the rule 2 is not passed
        """
        parents = parents or []
        children = children or []
        if parents:
            if any([p.undo for p in parents]) and not step.undo:
                raise errors.ParentUndoError(self.step.id, [p.undo for p in parents if p.undo])
        if children:
            if any([not c.undo for c in children]) and step.undo:
                raise errors.ChildDoError(self.step.id, [c.undo for c in children if not c.undo])
        g = self._graph.copy()
        g.add_edges_from([(p, step) for p in parents])
        g.add_edges_from([(step, c) for c in children])
        if g.is_cyclic():
            raise errors.CycleError()

    def add_step(self, *args, parents=None, children=None, **kwargs) -> Step:
        """Allows to add step into the orchestration

        :param args: args passed to Step: undo, action_template.
        :param parents: list of parent steps.
        :param children: list of children steps.
        :param kwargs: keyword arguments passed to the Step
        :return: The step created.
        """
        parents = parents or []
        children = children or []
        s = Step(None, *args, **kwargs)
        self._step_exists(parents + children)
        self._check_dependencies(s, parents, children)
        s.orchestration = self
        self._graph.add_node(s)
        self.add_parents(s, parents)
        self.add_children(s, children)
        return s

    def delete_step(self, step: Step) -> 'Orchestration':
        """
        Allows to remove a Step from the orchestration

        Parameters
        ----------
        step Step: step to remove from the orchestration
        """
        self._step_exists(step)
        i = self.steps.index(step)
        self.steps.pop(i)
        self._graph.remove_node(step)
        return self

    def add_parents(self, step: Step, parents: t.List[Step]) -> 'Orchestration':
        """add_parents adds the parents passed into the step. No remove from previous parents

        Parameters
        ----------
        step: Step
            step to add parents
        parents: list
            list of parent steps to add

        See Also
        --------
        set_parents, delete_parents
        add_children, set_children, delete_children

        Examples
        --------
        >>> at = ActionTemplate(name='action', version=1, action_type=ActionType.SHELL, code='code to run',
                                parameters={'param1': 'test'}, expected_output='',
                                expected_rc=0, system_kwargs={})
        >>> o = Orchestration('Test Orchestration', 1, DAG(), 'description')
        >>> s1 = o.add_step(undo=False, action_template=at, parents=[], children=[], stop_on_error=False)
        >>> s2 = o.add_step(undo=False, action_template=at, parents=[], children=[], stop_on_error=False)
        >>> s3 = o.add_step(undo=False, action_template=at, parents=[], children=[], stop_on_error=False)
        >>> o.add_parents(s2, [s1])
        >>> o.parents[s2]
        [Step1]
        >>> o.add_parents(s2, [s3])
        >>> o.parents[s2]
        [Step1, Step3]
        """
        self._step_exists([step] + list(parents))
        self._check_dependencies(step, parents=parents)
        step._add_parents(parents)
        self._graph.add_edges_from([(p, step) for p in parents])
        return self

    def delete_parents(self, step: Step, parents: t.List[Step]) -> 'Orchestration':
        """delete_parents deletes the parents passed from the step.

        Parameters
        ----------
        step: Step
            step to remove parents
        parents: list
            list of parent steps to remove

        See Also
        --------
        add_parents, set_parents
        add_children, set_children, delete_children

        Examples
        --------
        >>> at = ActionTemplate(name='action', version=1, action_type=ActionType.SHELL, code='code to run',
                                parameters={'param1': 'test'}, expected_output='',
                                expected_rc=0, system_kwargs={})
        >>> o = Orchestration('Test Orchestration', 1, DAG(), 'description')
        >>> s1 = o.add_step(undo=False, action_template=at, parents=[], children=[], stop_on_error=False)
        >>> s2 = o.add_step(undo=False, action_template=at, parents=[], children=[], stop_on_error=False)
        >>> s3 = o.add_step(undo=False, action_template=at, parents=[], children=[], stop_on_error=False)
        >>> o.add_children(s1, [s2, s3])
        >>> o.children[s1]
        [Step2, Step3]
        >>> o.delete_parents(s3, [s1])
        >>> o.children[s1]
        [Step2]
        """
        self._step_exists([step] + list(parents))
        step._remove_parents(parents)
        self._graph.remove_edges_from([(p, step) for p in parents])
        return self

    def set_parents(self, step: Step, parents: t.List[Step]) -> 'Orchestration':
        """set_parents sets the parents passed on the step, removing the previos ones

        Parameters
        ----------
        step: Step
            step to remove parents
        parents: list
            list of parent steps to set

        See Also
        --------
        add_parents, delete_parents
        add_children, delete_children, set_children,

        Examples
        --------
        >>> at = ActionTemplate(name='action', version=1, action_type=ActionType.SHELL, code='code to run',
                                parameters={'param1': 'test'}, expected_output='',
                                expected_rc=0, system_kwargs={})
        >>> o = Orchestration('Test Orchestration', 1, DAG(), 'description')
        >>> s1 = o.add_step(undo=False, action_template=at, parents=[], children=[], stop_on_error=False)
        >>> s2 = o.add_step(undo=False, action_template=at, parents=[], children=[], stop_on_error=False)
        >>> s3 = o.add_step(undo=False, action_template=at, parents=[], children=[], stop_on_error=False)
        >>> o.add_parents(s1, [s2])
        >>> o.parents[s1]
        [Step2]
        >>> o.set_parents(s1, [s3])
        >>> o.parents[s1]
        [Step3]
        """
        self.delete_parents(step, self._graph.pred[step])
        self.add_parents(step, parents)
        return self

    def add_children(self, step: Step, children: t.List[Step]) -> 'Orchestration':
        self._step_exists([step] + children)
        self._check_dependencies(step, children=children)
        step._add_children(children)
        self._graph.add_edges_from([(step, c) for c in children])
        return self

    def delete_children(self, step: Step, children: t.List[Step]) -> 'Orchestration':
        self._step_exists([step] + children)
        step._remove_children(children)
        self._graph.remove_edges_from([(step, c) for c in children])
        return self

    def set_children(self, step: Step, children: t.List[Step]) -> 'Orchestration':
        self.delete_children(step, self._graph.succ[step])
        self.add_children(step, children)
        return self

    def eq_imp(self, other: 'Orchestration') -> bool:
        """
        compares if two orchestrations implement same steps with same parameters and dependencies

        Parameters
        ----------
        other: Orchestration

        Returns
        -------
        result: bool
        """
        if isinstance(other, self.__class__):
            if len(self.steps) != len(other.steps):
                return False
            res = []
            for s in self.steps:
                res.append(any(map(lambda x: s.eq_imp(x), other.steps)))

            if all(res):

                matched_steps = []
                res2 = []
                v2 = []
                for k1, v1 in self.children.items():
                    k2 = None
                    for s in filter(lambda x: x not in matched_steps, other.steps):
                        if k1.eq_imp(s):
                            k2 = s
                            v2 = other.children[k2]
                            break
                    if not k2:
                        raise RuntimeError('Step not found in other')
                    matched_steps.append(k2)
                    if len(v1) != len(v2):
                        return False

                    for s in v1:
                        res2.append(any(map(lambda x: s.eq_imp(x), v2)))
                return all(res2)
            else:
                return False
        else:
            return False

    def subtree(self, steps: t.Union[t.List[Step], t.Iterable[Step]]) -> t.Dict[Step, t.List[Step]]:
        return self._graph.subtree(steps)

    def to_json(self, add_target=False, add_params=False, add_steps=False, add_action=False):
        data = super().to_json()
        data.update(name=self.name, version=self.version, stop_on_error=self.stop_on_error,
                    undo_on_error=self.undo_on_error, stop_undo_on_error=self.stop_undo_on_error)
        if add_target:
            data.update(target=list(self.target))
        if add_params:
            data.update(params=list(self.user_parameters))
        if add_steps:
            json_steps = []
            for step in self.steps:
                json_step = step.to_json(add_action=add_action)
                json_step.pop('orchestration_id')
                json_steps.append(json_step)
            data['steps'] = json_steps
        return data

    @classmethod
    def from_json(cls, kwargs):
        return super().from_json(kwargs)

    def __str__(self):
        return f"{self.name}.{self.version}"



import base64
import hashlib
import os
from unittest.mock import patch

from flask import url_for
from flask_jwt_extended import create_access_token
from pyfakefs.fake_filesystem_unittest import TestCase

from dm.domain.entities import Software, SoftwareServerAssociation, Server, Transfer, TransferStatus
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db


class TestTransferList(TestCase):

    def setUp(self) -> None:
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))

        db.create_all()
        set_initial()
        db.session.commit()

        self.source_path = '/software'
        self.filename = 'filename.zip'
        self.content = b'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
        self.size = len(self.content)
        self.checksum = hashlib.md5(self.content).hexdigest()

        self.server = Server.get_current()
        self.soft = Software(name='test_software', version=1, filename='software.zip', size=self.size,
                             checksum=self.checksum)

        self.ssa = SoftwareServerAssociation(software=self.soft, server=self.server,
                                             path=self.source_path)

        self.dest_path = '/dest_repo'

        db.session.add(self.soft)
        db.session.add(self.ssa)
        db.session.commit()

        self.setUpPyfakefs()
        self.fs.create_dir(self.source_path)
        self.fs.create_dir(self.dest_path)
        self.fs.create_file(os.path.join(self.source_path, self.filename), contents=self.content)
        self.fs.create_file(os.path.join(self.source_path, self.soft.filename), contents=self.content)

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()

    def test_get_transfers(self):
        t = Transfer(software=self.soft, dest_path=self.dest_path, num_chunks=16, status=TransferStatus.IN_PROGRESS)
        db.session.add(t)
        db.session.commit()
        resp = self.client.get(url_for('api_1_0.transferlist'), headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        self.assertListEqual([t.to_json()], resp.get_json())

    def test_post_create_software_transfer(self):
        self.fs.create_file(os.path.join('/new_dest_path', self.soft.filename + '_chunk.1'))
        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"software_id": str(self.soft.id),
                                      'dest_path': '/new_dest_path',
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(202, resp.status_code)
        data = resp.json
        t: Transfer = Transfer.query.get(data.get('id'))
        self.assertEqual(self.soft.id, t.software.id)
        self.assertEqual(16, t.num_chunks)
        self.assertEqual('/new_dest_path', t.dest_path)
        self.assertEqual(t.status, TransferStatus.WAITING_CHUNKS)

        self.assertFalse(os.path.exists(os.path.join('/new_dest_path', self.soft.filename + '_chunk.1')))

    def test_post_create_software_transfer_with_chunks_already_inside(self):
        self.fs.create_file(os.path.join(self.dest_path, self.soft.filename + '_chunk.1'))
        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"software_id": str(self.soft.id),
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(202, resp.status_code)
        data = resp.json
        t: Transfer = Transfer.query.get(data.get('id'))
        self.assertEqual(self.soft.id, t.software.id)
        self.assertEqual(16, t.num_chunks)
        self.assertEqual(self.dest_path, t.dest_path)
        self.assertEqual(t.status, TransferStatus.WAITING_CHUNKS)

        self.assertFalse(os.path.exists(os.path.join(self.dest_path, self.soft.filename + '_chunk.1')))

    def test_post_create_software_transfer_with_pending_transfer(self, ):
        t = Transfer(software=self.soft, dest_path=self.dest_path, num_chunks=16, status=TransferStatus.IN_PROGRESS)
        db.session.add(t)
        db.session.commit()

        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"software_id": str(self.soft.id), 'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(409, resp.status_code)
        self.assertDictEqual({"error": f"There is already a transfer sending software {self.soft.id}"}, resp.get_json())

    def test_post_create_software_transfer_with_pending_transfer_cancel(self):
        t = Transfer(software=self.soft, dest_path=self.dest_path, num_chunks=16, status=TransferStatus.IN_PROGRESS)
        db.session.add(t)
        db.session.commit()

        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"software_id": str(self.soft.id), 'dest_path': self.dest_path,
                                      'num_chunks': 16, 'cancel_pending': True}, headers=self.auth.header)
        self.assertEqual(202, resp.status_code)
        self.assertEqual(TransferStatus.CANCELED, t.status)

    def test_post_create_filename_transfer_with_pending_transfer(self):
        t = Transfer(software=self.filename, size=self.size, checksum=self.checksum, dest_path=self.dest_path,
                     num_chunks=16, status=TransferStatus.IN_PROGRESS)
        db.session.add(t)
        db.session.commit()

        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"filename": self.filename,
                                      'size': self.size,
                                      'checksum': self.checksum,
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(409, resp.status_code)
        self.assertDictEqual(
            {"error": f"There is already a transfer sending file {os.path.join(self.dest_path, self.filename)}"},
            resp.get_json())

    def test_post_create_filename_transfer_with_pending_transfer_cancel(self):
        t = Transfer(software=self.filename, size=self.size, checksum=self.checksum, dest_path=self.dest_path,
                     num_chunks=16, status=TransferStatus.IN_PROGRESS)
        db.session.add(t)
        db.session.commit()

        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"filename": self.filename,
                                      'size': self.size,
                                      'checksum': self.checksum,
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16,
                                      'cancel_pending': True}, headers=self.auth.header)
        self.assertEqual(202, resp.status_code)
        self.assertEqual(TransferStatus.CANCELED, t.status)

    def test_post_filename_and_software_not_specified(self, ):
        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={'filename': self.filename,
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(404, resp.status_code)
        self.assertDictEqual({'error': 'size and checksum must be specified when a filename set'}, resp.get_json())

        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={'filename': self.filename,
                                      'size': self.size,
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(404, resp.status_code)
        self.assertDictEqual({'error': 'size and checksum must be specified when a filename set'}, resp.get_json())

        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={'filename': self.filename,
                                      'checksum': self.checksum,
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(404, resp.status_code)
        self.assertDictEqual({'error': 'size and checksum must be specified when a filename set'}, resp.get_json())

    def test_post_filename_without_size_or_checksum(self):
        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(404, resp.status_code)
        self.assertDictEqual({'error': 'Neither filename nor software specified for the transfer'}, resp.get_json())

    def test_post_file_exists(self):
        self.fs.create_file(os.path.join(self.dest_path, self.filename))
        self.fs.create_file(os.path.join(self.dest_path, self.soft.filename))

        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"filename": self.filename,
                                      'size': self.size,
                                      'checksum': self.checksum,
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(409, resp.status_code)
        self.assertDictEqual({"error": "File already exists. Use force=True if needed"}, resp.get_json())

        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"software_id": str(self.soft.id),
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(409, resp.status_code)
        self.assertDictEqual({"error": "File already exists. Use force=True if needed"}, resp.get_json())

    def test_post_file_exists_force(self):
        self.fs.create_file(os.path.join(self.dest_path, self.filename))
        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"filename": self.filename,
                                      'size': self.size,
                                      'checksum': self.checksum,
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16,
                                      'force': False}, headers=self.auth.header)
        self.assertEqual(409, resp.status_code)
        self.assertDictEqual({"error": "File already exists. Use force=True if needed"}, resp.get_json())

        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"filename": self.filename,
                                      'size': self.size,
                                      'checksum': self.checksum,
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16,
                                      'force': True}, headers=self.auth.header)
        self.assertEqual(202, resp.status_code)
        self.assertFalse(os.path.exists(os.path.join(self.dest_path, self.filename)))

        self.fs.create_file(os.path.join(self.dest_path, self.soft.filename))

        with patch('dm.web.api_1_0.resources.transfer.os.remove') as mock_remove:
            mock_remove.side_effect = PermissionError('Not allowed')

            resp = self.client.post(url_for('api_1_0.transferlist'),
                                    json={'software_id': str(self.soft.id),
                                          'dest_path': self.dest_path,
                                          'num_chunks': 16,
                                          'force': True}, headers=self.auth.header)
            self.assertEqual(500, resp.status_code)
            file = os.path.join(self.dest_path, self.soft.filename)
            self.assertDictEqual({"error": f"Unable to remove {file}: Not allowed"}, resp.get_json())

    def test_post_create_transfer_error_with_pending_transfer(self):
        t = Transfer(software=self.soft, dest_path=self.dest_path, num_chunks=16, status=TransferStatus.IN_PROGRESS)
        db.session.add(t)
        db.session.commit()

        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"software_id": str(self.soft.id), 'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(409, resp.status_code)
        self.assertDictEqual({"error": f"There is already a transfer sending software {self.soft.id}"}, resp.get_json())

    @patch('dm.web.api_1_0.resources.transfer.os.makedirs', autospec=True)
    def test_post_create_software_error_on_create_dest_folder(self, mock_makedirs):
        mock_makedirs.side_effect = PermissionError('Not allowed')
        resp = self.client.post(url_for('api_1_0.transferlist'),
                                json={"software_id": str(self.soft.id),
                                      'dest_path': self.dest_path,
                                      'num_chunks': 16}, headers=self.auth.header)
        self.assertEqual(500, resp.status_code)
        self.assertDictEqual({'error': f"Unable to create dest path {self.dest_path}: Not allowed"}, resp.get_json())


class TestTransferResource(TestCase):

    def setUp(self) -> None:
        self.setUpPyfakefs()
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))

        db.create_all()
        set_initial()
        db.session.commit()

        self.source_path = '/software'
        self.filename = 'filename.zip'
        self.content = b'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
        self.size = len(self.content)
        self.checksum = hashlib.md5(self.content).hexdigest()
        self.dest_path = '/dest_repo'

        self.fs.create_dir(self.source_path)
        self.fs.create_dir(self.dest_path)
        self.fs.create_file(os.path.join(self.source_path, self.filename), contents=self.content)

        self.server = Server.get_current()
        self.transfer = Transfer(software=self.filename, size=self.size, checksum=self.checksum,
                                 dest_path=self.dest_path,
                                 num_chunks=16, status=TransferStatus.WAITING_CHUNKS)
        db.session.add(self.transfer)
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()

    def test_get_transfer(self):
        resp = self.client.get(url_for('api_1_0.transferresource', transfer_id=str(self.transfer.id)),
                               headers=self.auth.header)
        self.assertEqual(200, resp.status_code)
        self.assertDictEqual(self.transfer.to_json(), resp.get_json())

        resp = self.client.get(url_for('api_1_0.transferresource', transfer_id='aaaaaaaa-1234-5678-1234-56781234aaa1'),
                               headers=self.auth.header)
        self.assertEqual(404, resp.status_code)

    def test_post_transfer(self):
        self.assertIsNone(self.transfer.started_on)
        resp = self.client.post(url_for('api_1_0.transferresource', transfer_id=str(self.transfer.id)),
                                json={'chunk': 0,
                                      'content': base64.b64encode(b'abcd').decode('ascii')},
                                headers=self.auth.header)

        self.assertEqual(201, resp.status_code)
        db.session.refresh(self.transfer)
        self.assertDictEqual({'message': f"Chunk 0 from transfer {str(self.transfer.id)} generated successfully"},
                             resp.get_json())
        self.assertEqual(TransferStatus.IN_PROGRESS, self.transfer.status)
        self.assertIsNotNone(self.transfer.started_on)
        self.assertTrue(os.path.exists(os.path.join(self.dest_path, f"{self.filename}_chunk.0")))

    def test_post_create_filename_one_chunk(self):
        self.transfer.num_chunks = 1
        db.session.commit()
        resp = self.client.post(url_for('api_1_0.transferresource', transfer_id=str(self.transfer.id)),
                                json={'chunk': 1,
                                      'content': base64.b64encode(self.content).decode('ascii')},
                                headers=self.auth.header)
        self.assertEqual(201, resp.status_code)
        self.assertDictEqual(
            {'message': f"File {self.filename} from transfer {str(self.transfer.id)} generated successfully"},
            resp.get_json())
        self.assertTrue(os.path.exists(os.path.join(self.dest_path, self.filename)))

    def test_put_transfer_file(self):
        self.transfer.status = TransferStatus.IN_PROGRESS
        db.session.commit()
        for chunk_content, chunk_id in zip([self.content[i:i + 4] for i in range(0, len(self.content), 4)],
                                           range(0, 16)):
            self.fs.create_file(os.path.join(self.dest_path, f"{self.filename}_chunk.{chunk_id}"))
            with open(os.path.join(self.dest_path, f"{self.filename}_chunk.{chunk_id}"), 'wb') as fh:
                fh.write(chunk_content)

        resp = self.client.put(url_for('api_1_0.transferresource', transfer_id=str(self.transfer.id)),
                                 headers=self.auth.header)
        self.assertEqual(201, resp.status_code)
        self.assertDictEqual({'message': f"File {os.path.join(self.dest_path, self.filename)} from transfer "
                                         f"{self.transfer.id} recived successfully"}, resp.get_json())
        self.assertTrue(os.path.exists(os.path.join(self.dest_path, self.filename)))
        with open(os.path.join(self.dest_path, self.filename), 'rb') as fh:
            self.assertEqual(self.content, fh.read())

    def test_put_transfer_status_completed(self):
        self.transfer.status = TransferStatus.COMPLETED
        db.session.commit()

        resp = self.client.put(url_for('api_1_0.transferresource', transfer_id=str(self.transfer.id)),
                                 headers=self.auth.header)
        self.assertEqual(410, resp.status_code)
        self.assertDictEqual({'error': 'Transfer has already completed'}, resp.get_json())

    def test_put_transfer_status_waiting_chunks(self):
        resp = self.client.put(url_for('api_1_0.transferresource', transfer_id=str(self.transfer.id)),
                                 headers=self.auth.header)
        self.assertEqual(406, resp.status_code)
        self.assertDictEqual({'error': 'Transfer still waiting for chunks'}, resp.get_json())

    def test_create_file_error_chunks(self):
        self.transfer.status = TransferStatus.IN_PROGRESS
        db.session.commit()
        # Generate put with files
        for chunk_content, chunk_id in zip([self.content[i:i + 4] for i in range(0, len(self.content), 4)],
                                           [0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]):
            self.fs.create_file(os.path.join(self.dest_path, f"{self.filename}_chunk.{chunk_id}"))
            with open(os.path.join(self.dest_path, f"{self.filename}_chunk.{chunk_id}"), 'wb') as fh:
                fh.write(chunk_content)

        resp = self.client.put(
            url_for('api_1_0.transferresource', transfer_id=str(self.transfer.id), _external=False),
            json={},
            headers=self.auth.header)

        self.assertEqual(404, resp.status_code)
        self.assertDictEqual({"error": f"Not enough chunks to generate the file"}, resp.json)

    def test_create_file_error_no_chunk(self):
        self.transfer.status = TransferStatus.IN_PROGRESS
        db.session.commit()
        resp = self.client.put(
            url_for('api_1_0.transferresource', transfer_id=str(self.transfer.id), _external=False),
            json={},
            headers=self.auth.header)

        self.assertEqual(404, resp.status_code)
        self.assertDictEqual({"error": f"Any chunk found on {self.dest_path}"}, resp.json)

    def test_error_checksum(self):
        self.transfer.status = TransferStatus.IN_PROGRESS
        db.session.commit()
        # Generate put with files
        content = b'abcdefghijklmnopqrstuvwxyzXXXDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
        for chunk_content, chunk_id in zip([content[i:i + 4] for i in range(0, len(self.content), 4)],
                                           range(0, 16)):
            self.fs.create_file(os.path.join(self.dest_path, f"{self.filename}_chunk.{chunk_id}"))
            with open(os.path.join(self.dest_path, f"{self.filename}_chunk.{chunk_id}"), 'wb') as fh:
                fh.write(chunk_content)

        resp = self.client.put(
            url_for('api_1_0.transferresource', transfer_id=str(self.transfer.id), _external=False),
            json={},
            headers=self.auth.header)

        self.assertEqual(404, resp.status_code)
        db.session.refresh(self.transfer)
        self.assertDictEqual(
            {"error": f"Error on transfer '{str(self.transfer.id)}': Checksum error"},
            resp.get_json())
        self.assertEqual(self.transfer.status, TransferStatus.CHECKSUM_ERROR)

    def test_error_size_file(self):
        self.transfer.status = TransferStatus.IN_PROGRESS
        db.session.commit()
        # Generate put with files
        content = b'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ012345678'
        for chunk_content, chunk_id in zip([content[i:i + 4] for i in range(0, len(content), 4)],
                                           range(0, 16)):
            self.fs.create_file(os.path.join(self.dest_path, f"{self.filename}_chunk.{chunk_id}"))
            with open(os.path.join(self.dest_path, f"{self.filename}_chunk.{chunk_id}"), 'wb') as fh:
                fh.write(chunk_content)

        resp = self.client.put(
            url_for('api_1_0.transferresource', transfer_id=str(self.transfer.id), _external=False),
            json={},
            headers=self.auth.header)

        self.assertEqual(404, resp.status_code)
        db.session.refresh(self.transfer)
        self.assertDictEqual(
            {"error": f"Error on transfer '{str(self.transfer.id)}': Final file size does not match expected size"},
            resp.get_json())
        self.assertEqual(self.transfer.status, TransferStatus.SIZE_ERROR)



import concurrent.futures
import random
import time
from unittest import TestCase

from flask import Flask, current_app, g, request

from dm.web.extensions.flask_executor.executor import Executor, propagate_exceptions_callback


def fib(n):
    if n <= 2:
        return 1
    else:
        return fib(n - 1) + fib(n - 2)


def app_context_test_value(_=None):
    return current_app.config['TEST_VALUE']


def request_context_test_value(_=None):
    return request.test_value


def g_context_test_value(_=None):
    return g.test_value


def fail():
    time.sleep(0.1)
    raise RuntimeError


class Test(TestCase):
    def setUp(self) -> None:
        self.app = Flask(__name__)
        
    def test_init(self):
        self.app = Flask(__name__)
        executor = Executor(self.app)
        assert 'executor' in self.app.extensions
        assert isinstance(executor, concurrent.futures._base.Executor)
        assert isinstance(executor._self, concurrent.futures._base.Executor)
        assert getattr(executor, 'shutdown')

    def test_factory_init(self):
        self.app = Flask(__name__)
        executor = Executor()
        executor.init_app(self.app)
        assert 'executor' in self.app.extensions
        assert isinstance(executor._self, concurrent.futures._base.Executor)

    def test_thread_executor_init(self):
        self.app = Flask(__name__)
        self.app.config['EXECUTOR_TYPE'] = 'thread'
        executor = Executor(self.app)
        assert isinstance(executor._self, concurrent.futures.ThreadPoolExecutor)
        assert isinstance(executor, concurrent.futures.ThreadPoolExecutor)

    def test_process_executor_init(self):
        self.app = Flask(__name__)
        self.app.config['EXECUTOR_TYPE'] = 'process'
        executor = Executor(self.app)
        assert isinstance(executor._self, concurrent.futures.ProcessPoolExecutor)
        assert isinstance(executor, concurrent.futures.ProcessPoolExecutor)

    def test_default_executor_init(self):
        self.app = Flask(__name__)
        executor = Executor(self.app)
        assert isinstance(executor._self, concurrent.futures.ThreadPoolExecutor)

    def test_invalid_executor_init(self):
        self.app = Flask(__name__)
        self.app.config['EXECUTOR_TYPE'] = 'invalid_value'
        try:
            executor = Executor(self.app)
        except ValueError:
            assert True
        else:
            assert False

    def test_submit(self):
        self.app = Flask(__name__)
        executor = Executor(self.app)
        with self.app.test_request_context(''):
            future = executor.submit(fib, 5)
        assert future.result() == fib(5)

    def test_max_workers(self):
        self.app = Flask(__name__)
        EXECUTOR_MAX_WORKERS = 10
        self.app.config['EXECUTOR_MAX_WORKERS'] = EXECUTOR_MAX_WORKERS
        executor = Executor(self.app)
        assert executor._max_workers == EXECUTOR_MAX_WORKERS
        assert executor._self._max_workers == EXECUTOR_MAX_WORKERS

    def test_thread_decorator_submit(self):
        
        self.app.config['EXECUTOR_TYPE'] = 'thread'
        executor = Executor(self.app)

        @executor.job
        def decorated(n):
            return fib(n)

        with self.app.test_request_context(''):
            future = decorated.submit(5)
        assert future.result() == fib(5)

    def test_thread_decorator_submit_stored(self):
        self.app.config['EXECUTOR_TYPE'] = 'thread'
        executor = Executor(self.app)

        @executor.job
        def decorated(n):
            return fib(n)

        with self.app.test_request_context():
            future = decorated.submit_stored('fibonacci', 35)
        assert executor.futures.done('fibonacci') is False
        assert future in executor.futures
        executor.futures.pop('fibonacci')
        assert future not in executor.futures

    def test_thread_decorator_map(self):
        iterable = list(range(5))
        self.app.config['EXECUTOR_TYPE'] = 'thread'
        executor = Executor(self.app)

        @executor.job
        def decorated(n):
            return fib(n)

        with self.app.test_request_context(''):
            results = decorated.map(iterable)
        for i, r in zip(iterable, results):
            assert fib(i) == r

    def test_process_decorator(self):
        ''' Using decorators should fail with a TypeError when using the ProcessPoolExecutor '''
        self.app.config['EXECUTOR_TYPE'] = 'process'
        executor = Executor(self.app)
        try:
            @executor.job
            def decorated(n):
                return fib(n)
        except TypeError:
            pass
        else:
            assert 0

    def test_submit_app_context(self):
        test_value = random.randint(1, 101)
        self.app.config['TEST_VALUE'] = test_value
        executor = Executor(self.app)
        with self.app.test_request_context(''):
            future = executor.submit(app_context_test_value)
        assert future.result() == test_value

    def test_submit_g_context_process(self):
        test_value = random.randint(1, 101)
        executor = Executor(self.app)
        with self.app.test_request_context(''):
            g.test_value = test_value
            future = executor.submit(g_context_test_value)
        assert future.result() == test_value

    def test_submit_request_context(self):
        test_value = random.randint(1, 101)
        executor = Executor(self.app)
        with self.app.test_request_context(''):
            request.test_value = test_value
            future = executor.submit(request_context_test_value)
        assert future.result() == test_value

    def test_map_app_context(self):
        test_value = random.randint(1, 101)
        iterator = list(range(5))
        self.app.config['TEST_VALUE'] = test_value
        executor = Executor(self.app)
        with self.app.test_request_context(''):
            results = executor.map(app_context_test_value, iterator)
        for r in results:
            assert r == test_value

    def test_map_g_context_process(self):
        test_value = random.randint(1, 101)
        iterator = list(range(5))
        executor = Executor(self.app)
        with self.app.test_request_context(''):
            g.test_value = test_value
            results = executor.map(g_context_test_value, iterator)
        for r in results:
            assert r == test_value

    def test_map_request_context(self):
        test_value = random.randint(1, 101)
        iterator = list(range(5))
        executor = Executor(self.app)
        with self.app.test_request_context('/'):
            request.test_value = test_value
            results = executor.map(request_context_test_value, iterator)
        for r in results:
            assert r == test_value

    def test_executor_stored_future(self):
        executor = Executor(self.app)
        with self.app.test_request_context():
            future = executor.submit_stored('fibonacci', fib, 35)
        assert executor.futures.done('fibonacci') is False
        assert future in executor.futures
        executor.futures.pop('fibonacci')
        assert future not in executor.futures

    def test_set_max_futures(self):
        self.app.config['EXECUTOR_FUTURES_MAX_LENGTH'] = 10
        executor = Executor(self.app)
        assert executor.futures.max_length == self.app.config['EXECUTOR_FUTURES_MAX_LENGTH']

    def test_named_executor(self):
        name = 'custom'
        EXECUTOR_MAX_WORKERS = 5
        CUSTOM_EXECUTOR_MAX_WORKERS = 10
        self.app.config['EXECUTOR_MAX_WORKERS'] = EXECUTOR_MAX_WORKERS
        self.app.config['CUSTOM_EXECUTOR_MAX_WORKERS'] = CUSTOM_EXECUTOR_MAX_WORKERS
        executor = Executor(self.app)
        custom_executor = Executor(self.app, name=name)
        assert 'executor' in self.app.extensions
        assert name + 'executor' in self.app.extensions
        assert executor._self._max_workers == EXECUTOR_MAX_WORKERS
        assert executor._max_workers == EXECUTOR_MAX_WORKERS
        assert custom_executor._self._max_workers == CUSTOM_EXECUTOR_MAX_WORKERS
        assert custom_executor._max_workers == CUSTOM_EXECUTOR_MAX_WORKERS

    def test_named_executor_submit(self):
        name = 'custom'
        custom_executor = Executor(self.app, name=name)
        with self.app.test_request_context(''):
            future = custom_executor.submit(fib, 5)
        assert future.result() == fib(5)

    def test_named_executor_name(self):
        name = 'invalid name'
        try:
            executor = Executor(self.app, name=name)
        except ValueError:
            assert True
        else:
            assert False

    def test_default_done_callback(self):
        executor = Executor(self.app)

        def callback(future):
            setattr(future, 'test', 'test')

        executor.add_default_done_callback(callback)
        with self.app.test_request_context('/'):
            future = executor.submit(fib, 5)
            concurrent.futures.wait([future])
            assert hasattr(future, 'test')

    def test_propagate_exception_callback(self):
        self.app.config['EXECUTOR_PROPAGATE_EXCEPTIONS'] = True
        executor = Executor(self.app)
        with self.assertRaises(RuntimeError):
            with self.app.test_request_context('/'):
                future = executor.submit(fail)
                assert propagate_exceptions_callback in future._done_callbacks
                concurrent.futures.wait([future])
                propagate_exceptions_callback(future)

    def test_coerce_config_types(self):
        self.app.config['EXECUTOR_MAX_WORKERS'] = '5'
        self.app.config['EXECUTOR_FUTURES_MAX_LENGTH'] = '10'
        self.app.config['EXECUTOR_PROPAGATE_EXCEPTIONS'] = 'true'
        executor = Executor(self.app)
        with self.app.test_request_context():
            future = executor.submit_stored('fibonacci', fib, 35)

    def test_shutdown_executor(self):
        executor = Executor(self.app)
        assert executor._shutdown is False
        executor.shutdown()
        assert executor._shutdown is True

    def test_pre_init_executor(self):
        executor = Executor()

        @executor.job
        def decorated(n):
            return fib(n)

        assert executor
        executor.init_app(self.app)
        with self.app.test_request_context(''):
            future = decorated.submit(5)
        assert future.result() == fib(5)



# If the given text doesn't end with a newline, the interface won't finish.
from unittest import TestCase

from commands import nested_dict
from completer import DshellCompleter
from prompt_toolkit import PromptSession
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.output import DummyOutput


class TestCompleter(TestCase):

    def test_completer(self):
        text = 'server \t\r'

        inp = create_pipe_input()

        try:
            inp.send_text(text)
            session = PromptSession(
                input=inp,
                output=DummyOutput(),
                editing_mode=EditingMode.VI,
                completer=DshellCompleter.from_nested_dict(nested_dict)
            )

            result = session.prompt()
        finally:
            inp.close()


from unittest import TestCase
from unittest.mock import patch, MagicMock

import requests
import responses
from flask import Flask

from dm.domain.entities import Server, Route
from dm.web import db, errors
from dm.web.decorators import forward_or_dispatch
from tests.helpers import ValidateResponseMixin


class TestForwardOrDispatch(TestCase, ValidateResponseMixin):

    # def run(self, result=None):
    #     with patch('dm.web.scopefunc') as mock_scopefunc:
    #         mock_scopefunc.side_effect = lambda: threading.get_ident()
    #         super().run(result)

    def setUp(self):
        """Create and configure a new self.app instance for each test."""
        # create a temporary file to isolate the database for each test
        # create the self.app with common test config
        self.app = Flask(__name__)

        @self.app.route('/', methods=['GET', 'POST'])
        @forward_or_dispatch
        def hello():
            return {'msg': 'default response'}
        
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.init_app(self.app)
        db.create_all()

        self.srv1 = Server(id='bbbbbbbb-1234-5678-1234-56781234bbb1', name='server1',
                           dns_or_ip='192.168.1.9', port=7123)
        Route(self.srv1, cost=0)
        self.srv2 = Server(id='bbbbbbbb-1234-5678-1234-56781234bbb2', name='server2',
                           dns_or_ip='192.168.1.10', port=7124)
        Route(self.srv2, proxy_server=self.srv1, cost=1)
        db.session.add_all([self.srv1, self.srv2])
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @patch('dm.web.decorators.g')
    @responses.activate
    def test_forward_or_dispatch_in_content(self, mock_g):
        mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')
        responses.add(responses.POST, 'http://192.168.1.9:7123/', json={'msg': 'response'})

        # check if request is forwarded to the server
        response = self.client.post('/', json={'destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2', 'data': None})
        self.assertEqual({'msg': 'response'}, response.json)

        self.assertEqual(1, len(responses.calls))
        self.assertEqual('http://192.168.1.9:7123/', responses.calls[0].request.url)

        # check if request is forwarded to the server
        response = self.client.post('/', json={'destination': 'bbbbbbbb-1234-5678-1234-56781234bbb1', 'data': None})
        self.assertEqual({'msg': 'default response'}, response.json)

        self.assertEqual(1, len(responses.calls))


    @patch('dm.web.decorators.g')
    @responses.activate
    def test_forward_or_dispatch(self, mock_g):
        mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')
        responses.add(responses.POST, 'http://192.168.1.9:7123/', json={'msg': 'response'})

        # check if request is forwarded to the server
        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2'})
        self.assertEqual({'msg': 'response'}, response.json)

        self.assertEqual(1, len(responses.calls))
        self.assertEqual('http://192.168.1.9:7123/', responses.calls[0].request.url)

        # check if request is forwarded to the server
        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb1'})
        self.assertEqual({'msg': 'default response'}, response.json)

        self.assertEqual(1, len(responses.calls))

    @patch('dm.web.decorators.g')
    def test_forward_or_dispatch_hidden_ip(self, mock_g):
        mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')

        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb1',
                                             'D-Source': 'bbbbbbbb-1234-5678-1234-56781234bbb2'},
                                    environ_base={'REMOTE_ADDR': '10.1.2.3'})
        self.assertEqual({'msg': 'default response'}, response.json)

        self.assertEqual(1, len(self.srv2.hidden_gates))
        hg = self.srv2.hidden_gates[0]
        self.assertEqual('10.1.2.3', str(hg.ip))
        self.assertEqual(7124, hg.port)

        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb1',
                                             'D-Source': 'bbbbbbbb-1234-5678-1234-56781234bbb2'},
                                    environ_base={'REMOTE_ADDR': '10.1.2.3'})
        self.assertEqual({'msg': 'default response'}, response.json)

        self.assertEqual(1, len(self.srv2.hidden_gates))
        self.assertEqual('10.1.2.3', str(hg.ip))
        self.assertEqual(7124, hg.port)


        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb1',
                                             'D-Source': 'bbbbbbbb-1234-5678-1234-56781234bbb2'},
                                    environ_base={'REMOTE_ADDR': '10.1.2.4'})
        self.assertEqual({'msg': 'default response'}, response.json)

        db.session.refresh(self.srv2)
        self.assertEqual(1, len(self.srv2.hidden_gates))
        self.assertEqual('10.1.2.4', str(hg.ip))
        self.assertEqual(7124, hg.port)


    @patch('dm.web.decorators.g')
    def test_forward_or_dispatch_hidden_ip_multiple_ports(self, mock_g):
        mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')

        g = self.srv2.add_new_gate('127.0.0.1', 7125)

        db.session.add(g)
        db.session.commit()

        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb1',
                                             'D-Source': 'bbbbbbbb-1234-5678-1234-56781234bbb2'},
                                    environ_base={'REMOTE_ADDR': '10.1.2.3'})

        self.assertEqual({'msg': 'default response'}, response.json)

        db.session.refresh(self.srv2)

        self.assertEqual(2, len(self.srv2.hidden_gates))
        hg = self.srv2.hidden_gates
        self.assertEqual('10.1.2.3', str(hg[0].ip))
        self.assertEqual(7124, hg[0].port)
        self.assertEqual('10.1.2.3', str(hg[1].ip))
        self.assertEqual(7125, hg[1].port)

        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb1',
                                             'D-Source': 'bbbbbbbb-1234-5678-1234-56781234bbb2'},
                                    environ_base={'REMOTE_ADDR': '10.1.2.3'})
        self.assertEqual({'msg': 'default response'}, response.json)

        db.session.refresh(self.srv2)

        self.assertEqual(2, len(self.srv2.hidden_gates))
        hg = self.srv2.hidden_gates
        self.assertEqual('10.1.2.3', str(hg[0].ip))
        self.assertEqual(7124, hg[0].port)
        self.assertEqual('10.1.2.3', str(hg[1].ip))
        self.assertEqual(7125, hg[1].port)

        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb1',
                                             'D-Source': 'bbbbbbbb-1234-5678-1234-56781234bbb2'},
                                    environ_base={'REMOTE_ADDR': '10.1.2.4'})

        self.assertEqual({'msg': 'default response'}, response.json)

        db.session.refresh(self.srv2)

        self.assertEqual(2, len(self.srv2.hidden_gates))
        hg = self.srv2.hidden_gates
        self.assertEqual('10.1.2.4', str(hg[0].ip))
        self.assertIn(hg[0].port, (7124, 7125))
        self.assertEqual('10.1.2.4', str(hg[1].ip))
        self.assertIn(hg[1].port, (7124, 7125))

    @patch('dm.web.decorators.g')
    def test_forward_or_dispatch_using_proxy_server(self, mock_g):
        me = Server(id='bbbbbbbb-1234-5678-1234-56781234bbb0', name='server0',
                    dns_or_ip='192.168.1.8', port=7123, me=True)
        db.session.add(me)
        db.session.commit()
        mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb0')

        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb0',
                                             'D-Source': 'bbbbbbbb-1234-5678-1234-56781234bbb2:bbbbbbbb-1234-5678-1234-56781234bbb1'},
                                    environ_base={'REMOTE_ADDR': '192.168.1.9'})
        self.assertEqual({'msg': 'default response'}, response.json)

        db.session.refresh(self.srv2)
        self.assertEqual(0, len(self.srv2.hidden_gates))

        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb0',
                                             'D-Source': 'bbbbbbbb-1234-5678-1234-56781234bbb2:bbbbbbbb-1234-5678-1234-56781234bbb1'},
                                    environ_base={'REMOTE_ADDR': '10.1.2.3'})
        self.assertEqual({'msg': 'default response'}, response.json)

        db.session.refresh(self.srv1)
        self.assertEqual(1, len(self.srv1.hidden_gates))
        hg = self.srv1.hidden_gates[0]
        self.assertEqual('10.1.2.3', str(hg.ip))
        self.assertEqual(7123, hg.port)

    @patch('dm.web.decorators.g')
    def test_server_not_found(self, mock_g):
        mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')

        resp = self.client.post('/', json={'data': None},
                                headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb5'})

        self.validate_error_response(resp, errors.EntityNotFound('Server', 'bbbbbbbb-1234-5678-1234-56781234bbb5'))

    @patch('dm.web.decorators.socket.gethostbyname')
    @patch('dm.web.decorators.g')
    def test_forward_or_dispatch_dns(self, mock_g, mock_gethostbyname):
        mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')

        def gethostbyname(dns: str):
            if dns == 'server2':
                return '10.1.2.3'
            else:
                return None

        mock_gethostbyname.side_effect = gethostbyname
        self.srv2.add_new_gate('server2', 7124)
        db.session.commit()

        response = self.client.post('/', json={'data': None},
                                    headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb1',
                                             'D-Source': 'bbbbbbbb-1234-5678-1234-56781234bbb2'},
                                    environ_base={'REMOTE_ADDR': '10.1.2.3'})
        self.assertEqual({'msg': 'default response'}, response.json)

        self.assertEqual(0, len(self.srv2.hidden_gates))

    @patch('dm.web.decorators.requests.request', autospec=True)
    @patch('dm.web.decorators.g')
    def test_forward_or_dispatch_proxy_request(self, mock_g, mock_request):
        mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')

        def request(*args, **kwargs):
            r = requests.Response()
            r._content = kwargs['headers'].get('d-source')
            r.status_code = 222
            return r

        mock_request.side_effect = request

        # check if request is forwarded to the server without d-source header
        resp = self.client.post('/', json={'data': None},
                                headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2'})

        self.assertEqual(':bbbbbbbb-1234-5678-1234-56781234bbb1', resp.get_data(True))

        # check if request is forwarded to the server with d-source header
        resp = self.client.post('/', json={'data': None},
                                headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2',
                                         'D-Source': 'bbbbbbbb-1234-5678-1234-56781234bbb0'})

        self.assertEqual('bbbbbbbb-1234-5678-1234-56781234bbb0:bbbbbbbb-1234-5678-1234-56781234bbb1',
                         resp.get_data(True))

    @patch('dm.web.decorators.requests.request', autospec=True)
    @patch('dm.web.decorators.g')
    def test_forward_or_dispatch_error_proxying(self, mock_g, mock_request):
        mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')
        mock_request.side_effect = requests.exceptions.ConnectionError('error')

        # check if request is forwarded to the server
        resp = self.client.post('/', json={'data': None},
                                headers={'D-Destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2'})

        self.validate_error_response(resp, errors.ProxyForwardingError(self.srv2,
                                                                       requests.exceptions.ConnectionError('error')))



from datetime import datetime
from unittest import TestCase, mock

from dm.use_cases.deployment import UndoCommand, Command
from dm.use_cases.operations import CompletedProcess


class TestCommand(TestCase):

    def setUp(self) -> None:
        self.mock_implementation = mock.Mock()
        self.mock_undo_command = mock.Mock()
        self.mock_parameters = mock.Mock()

    def test_invoke(self):
        with mock.patch('dm.use_cases.deployment.Command.success', new_callable=mock.PropertyMock) as mock_success:
            mock_success.return_value = True

            c = Command(implementation=self.mock_implementation, undo_command=self.mock_undo_command,
                        var_context=self.mock_parameters,
                        undo_on_error=False, id_=1)
            r = c.invoke()
            # call a second time to see if only one time is called
            r = c.invoke()
            self.mock_implementation.execute.assert_called_once_with(self.mock_parameters, timeout=None)
            self.assertTrue(r)

    def test_undo_command_succeed(self):
        with mock.patch('dm.use_cases.deployment.Command.success', new_callable=mock.PropertyMock) as mock_success:
            mock_success.return_value = True
            type(self.mock_undo_command).success = mock.PropertyMock(return_value=None)
            self.mock_undo_command.invoke.return_value = True

            c = Command(implementation=self.mock_implementation, undo_command=self.mock_undo_command,
                        var_context=self.mock_parameters,
                        undo_on_error=False, id_=1)
            r = c.undo()
            self.mock_undo_command.invoke.assert_called_once()
            self.assertTrue(r)

    def test_undo_command_not_succeed(self):
        with mock.patch('dm.use_cases.deployment.Command.success', new_callable=mock.PropertyMock) as mock_success:
            mock_success.return_value = False
            type(self.mock_undo_command).success = mock.PropertyMock(return_value=None)
            self.mock_undo_command.invoke.return_value = True

            c = Command(implementation=self.mock_implementation, undo_command=self.mock_undo_command,
                        var_context=self.mock_parameters,
                        undo_on_error=False, id_=1)
            r = c.undo()
            self.mock_undo_command.invoke.assert_not_called()
            self.assertIsNone(r)
            c.undo_on_error = True
            r = c.undo()
            self.mock_undo_command.invoke.assert_called_once()
            self.assertTrue(r)

    def test_undo_command_invoke_not_executed(self):
        with mock.patch('dm.use_cases.deployment.Command.success', new_callable=mock.PropertyMock) as mock_success:
            mock_success.return_value = None
            type(self.mock_undo_command).success = mock.PropertyMock(return_value=None)
            self.mock_undo_command.invoke.return_value = True

            c = Command(implementation=self.mock_implementation, undo_command=self.mock_undo_command,
                        var_context=self.mock_parameters,
                        undo_on_error=False, id_=1)
            r = c.undo()
            self.mock_undo_command.invoke.assert_not_called()
            self.assertIsNone(r)

    def test_result(self):
        type(self.mock_undo_command).success = mock.PropertyMock(return_value=True)
        self.mock_undo_command.invoke.return_value = True
        self.mock_undo_command._id.return_value = 2
        type(self.mock_undo_command).result = mock.PropertyMock(return_value={2: {'a': 2}})

        c = Command(implementation=self.mock_implementation, undo_command=self.mock_undo_command,
                    var_context=self.mock_parameters,
                    undo_on_error=False, id_=1)
        c._cp = {'success': True}
        self.assertDictEqual({1: {'success': True}, 2: {'a': 2}}, c.result)

    def test_fetch_result(self):
        self.mock_implementation.execute.return_value = CompletedProcess(success=True, stdout='output: var',
                                                                         stderr='',
                                                                         rc=0, start_time=datetime(2019, 4, 1))
        c = Command(implementation=self.mock_implementation, undo_command=self.mock_undo_command,
                    var_context=self.mock_parameters,
                    undo_on_error=False, post_process=r'output: (?P<output>\w+)', id_=1)
        r = c.invoke()
        self.assertTrue(r)
        self.mock_parameters.set.assert_called_once_with('output', 'var')

        self.mock_parameters.reset_mock()
        c = Command(implementation=self.mock_implementation, undo_command=self.mock_undo_command,
                    var_context=self.mock_parameters,
                    undo_on_error=False, post_process=r'output2: (?P<output>\w+)', id_=1)
        r = c.invoke()
        self.assertTrue(r)
        self.mock_parameters.set.assert_not_called()

        self.mock_implementation.execute.return_value = CompletedProcess(success=True, stdout='',
                                                                         stderr='output: var',
                                                                         rc=0, start_time=datetime(2019, 4, 1))
        self.mock_parameters.reset_mock()
        c = Command(implementation=self.mock_implementation, undo_command=self.mock_undo_command,
                    var_context=self.mock_parameters,
                    undo_on_error=False, post_process=r'output: (?P<output>\w+)', id_=1)
        r = c.invoke()
        self.assertTrue(r)
        self.mock_parameters.set.assert_called_with('output', 'var')

    def test_command_no_undo(self):
        cp = CompletedProcess(success=True, stdout='output: var',
                              stderr='',
                              rc=0, start_time=datetime(2019, 4, 1))
        self.mock_implementation.execute.return_value = cp
        c = Command(implementation=self.mock_implementation, var_context={'param': 'a'},
                    undo_on_error=False, id_=1)

        r = c.invoke()
        self.assertTrue(r)
        r = c.undo()
        self.assertTrue(r)
        self.assertDictEqual({1: cp}, c.result)

    def test_command_post_process(self):
        cp = CompletedProcess(success=True, stdout='{"output": "this is a message"}',
                              stderr='',
                              rc=0, start_time=datetime(2019, 4, 1))
        self.mock_implementation.execute.return_value = cp
        c = Command(implementation=self.mock_implementation, var_context=self.mock_parameters,
                    undo_on_error=False, post_process="vc.set('response', json.loads(cp.stdout))", id_=1)

        c.invoke()
        self.mock_parameters.set.assert_called_once_with('response', {"output": "this is a message"})

        c = Command(implementation=self.mock_implementation, var_context=self.mock_parameters,
                    undo_on_error=False, post_process="raise RuntimeError()", id_=1)

        c.invoke()
        self.assertIsInstance(cp.pre_post_error, RuntimeError)
        self.assertFalse(cp.success)

    def test_command_register(self):
        cp = CompletedProcess(success=True, stdout='{"output": "this is a message"}',
                              stderr='',
                              rc=0, start_time=datetime(2019, 4, 1))
        self.mock_implementation.execute.return_value = cp
        mock_register = mock.Mock()
        c = Command(implementation=self.mock_implementation, var_context=self.mock_parameters, register=mock_register,
                    undo_on_error=False, post_process="", id_=1)

        c.invoke()

        mock_register.register_step_execution.assert_called_once_with(c)


class TestUndoCommand(TestCase):

    def setUp(self) -> None:
        self.mock_implementation = mock.Mock()

    def test_invoke(self):
        completed_process = mock.Mock()
        type(completed_process).success = mock.PropertyMock(return_value=True)
        self.mock_implementation.execute.return_value = completed_process
        uc = UndoCommand(implementation=self.mock_implementation, var_context={'param': 'a'}, id_=1)

        uc.invoke(timeout=10)
        r = uc.invoke()
        self.mock_implementation.execute.assert_called_once_with({'param': 'a'}, timeout=10)
        self.assertTrue(r)

    def test_undo(self):
        uc = UndoCommand(implementation=self.mock_implementation, var_context={'param': 'a'}, id_=1)

        r = uc.undo()
        self.assertTrue(r)

    def test_execution(self):
        uc = UndoCommand(implementation=self.mock_implementation, var_context={'param': 'a'}, id_=1)
        uc._cp = {'success': True}
        self.assertDictEqual({1: {'success': True}}, uc.result)



import copy
from enum import Enum

from sqlalchemy import orm

from dm import defaults
from dm.domain.entities.base import UUIDistributedEntityMixin
from dm.utils import typos
from dm.web import db


class ActionType(Enum):
    ANSIBLE = 1
    PYTHON = 2
    SHELL = 3
    ORCHESTRATION = 4
    REQUEST = 5
    NATIVE = 6


class ActionTemplate(db.Model, UUIDistributedEntityMixin):
    __tablename__ = 'D_action_template'
    order = 10
    name = db.Column(db.String(40), nullable=False)
    version = db.Column(db.Integer, nullable=False)
    action_type = db.Column(typos.Enum(ActionType), nullable=False)
    code = db.Column(db.Text, nullable=False)
    parameters = db.Column(db.JSON)
    expected_stdout = db.Column(db.Text)
    expected_stderr = db.Column(db.Text)
    expected_rc = db.Column(db.Integer)
    system_kwargs = db.Column(db.JSON)
    pre_process = db.Column(db.Text)
    post_process = db.Column(db.Text)

    def __init__(self, name: str, version: int, action_type: ActionType, code: str = None, parameters: typos.Kwargs = None,
                 expected_stdout: str = None, expected_stderr: str = None, expected_rc: int = None,
                 system_kwargs: typos.Kwargs = None, pre_process: str = None, post_process: str = None,
                 **kwargs):
        UUIDistributedEntityMixin.__init__(self, **kwargs)
        self.name = name
        self.version = version
        self.action_type = action_type
        self.code = code
        self.parameters = parameters or {}
        self.expected_stdout = expected_stdout
        self.expected_stderr = expected_stderr
        self.expected_rc = expected_rc
        self.system_kwargs = system_kwargs or {}
        self.pre_process = pre_process
        self.post_process = post_process

    __table_args__ = (db.UniqueConstraint('name', 'version'),)

    @orm.reconstructor
    def init_on_load(self):
        self.parameters = self.parameters or {}
        self.system_kwargs = self.system_kwargs or {}

    def to_json(self):
        data = super().to_json()
        data.update(name=self.name, version=self.version,
                    action_type=self.action_type.name,
                    code=self.code)
        data.update(parameters=self.parameters)
        data.update(system_kwargs=self.system_kwargs)
        data.update(expected_stdout=self.expected_stdout)
        data.update(expected_stderr=self.expected_stderr)
        data.update(expected_rc=self.expected_rc)
        data.update(pre_process=self.pre_process)
        data.update(post_process=self.post_process)
        return data

    @classmethod
    def from_json(cls, kwargs):
        kwargs = copy.deepcopy(kwargs)
        kwargs['action_type'] = ActionType[kwargs.get('action_type')]
        return super().from_json(kwargs)

    @classmethod
    def set_initial(cls):
        from dm.domain.entities import bypass_datamark_update
        with bypass_datamark_update():
            at = cls.query.get('00000000-0000-0000-000a-000000000001')
            if at is None:
                at = ActionTemplate(name='send', version=1, action_type=ActionType.REQUEST,
                                    code='{"method": "post",'\
                                         '"view":"api_1_0.send",'\
                                         '"json": {"software_id": "{{software_id}}", "dest_server_id": "{{server_id}}"'\
                                         '{% if dest_path is defined %}, "dest_path":"{{dest_path}}"{% endif %}'\
                                         '{% if chunk_size is defined %}, "chunk_size":"{{chunk_size}}"{% endif %}'\
                                         '{% if max_senders is defined %}, "max_senders":"{{max_senders}}"{% endif %}'\
                                         ', "background": false, "include_transfer_data": true, "force": true} }',
                                    expected_rc=201, last_modified_at=defaults.INITIAL_DATEMARK,
                                    id='00000000-0000-0000-000a-000000000001',
                                    post_process="json_data=json.loads(cp.stdout)\nvc.set('file', json_data.get('file'))")

                db.session.add(at)
            at = cls.query.get('00000000-0000-0000-000a-000000000002')
            if at is None:
                at = ActionTemplate(name='wait', version=1, action_type=ActionType.NATIVE,
                                    code='{{list_server_names}} {{timeout}}',
                                    last_modified_at=defaults.INITIAL_DATEMARK,
                                    id='00000000-0000-0000-000a-000000000002')
                db.session.add(at)
            at = cls.query.get('00000000-0000-0000-000a-000000000003')
            if at is None:
                at = ActionTemplate(name='orchestration', version=1, action_type=ActionType.ORCHESTRATION,
                                    code='{{orchestration_id}} {{hosts}}',
                                    last_modified_at=defaults.INITIAL_DATEMARK,
                                    id='00000000-0000-0000-000a-000000000003')
                db.session.add(at)



from flask import request
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities import Step
from dm.web import db
from dm.web.decorators import forward_or_dispatch, securizer, validate_schema, lock_catalog
from dm.web.json_schemas import step_children


class StepRelationshipChildren(Resource):
    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, step_id):
        s: Step = Step.query.get_or_404(step_id)
        return dict(child_step_ids=[str(cs.id) for cs in s.children]), 200

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(step_children)
    @lock_catalog
    def patch(self, step_id):
        s: Step = Step.query.get_or_404(step_id)
        child_step_ids = request.get_json()['child_step_ids']
        child_steps = []
        for child_step_id in child_step_ids:
            child_steps.append(Step.query.get_or_404(child_step_id))
        s.orchestration.set_children(s, child_steps)

        db.session.commit()

        return dict(child_step_ids=[str(cs.id) for cs in s.children]), 200

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(step_children)
    @lock_catalog
    def post(self, step_id):
        s = Step.query.get_or_404(step_id)
        child_step_ids = request.get_json()['child_step_ids']
        child_steps = []
        for child_step_id in child_step_ids:
            child_steps.append(Step.query.get_or_404(child_step_id))
        s.orchestration.add_children(s, child_steps)

        db.session.commit()

        return dict(child_step_ids=[str(cs.id) for cs in s.children]), 200

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(step_children)
    @lock_catalog
    def delete(self, step_id):
        s = Step.query.get_or_404(step_id)
        child_step_ids = request.get_json()['child_step_ids']
        child_steps = []
        for child_step_id in child_step_ids:
            child_steps.append(Step.query.get_or_404(child_step_id))

        s.orchestration.delete_children(s, child_steps)

        db.session.commit()

        return dict(child_step_ids=[str(cs.id) for cs in s.children]), 200


import threading
from unittest import TestCase

from dm.utils.talkback import Talkback


class TestTalkback(TestCase):

    def test_wait(self):
        t = Talkback()

        th = threading.Timer(0.001, t.update, kwargs=dict(key=1))
        th.start()
        self.assertTrue(t.wait_exists('key', timeout=10))
        self.assertEqual(0, len(t._listeners))
        self.assertDictEqual({'key': 1}, t)

        self.assertTrue(t.wait_exists('key', timeout=0.01))
        self.assertEqual(0, len(t._listeners))

        th = threading.Timer(0.001, t.update, args=((('a', 2),),))
        th.start()
        self.assertTrue(t.wait_exists('a', timeout=10))
        self.assertEqual(0, len(t._listeners))
        self.assertIn('a', t)

        th = threading.Timer(0.001, t.update, args=({'key': 3},))
        th.start()
        self.assertTrue(t.wait_update('key', timeout=10))
        self.assertEqual(0, len(t._listeners))
        self.assertDictEqual({'key': 3, 'a': 2}, t)

        self.assertFalse(t.wait_exists('none', timeout=0.01))



import re
import typing as t
from collections import ChainMap


class VarContext:
    _exp = re.compile(r'\{\{\s*([\.\w]+)\s*\}\}')
    __reserved_keys = ('execution_server_id',)

    def __init__(self, globals: t.Mapping[str, t.Any] = None, initials: t.Mapping[str, t.Any] = None,
                 defaults: t.Mapping[str, t.Any] = None,
                 variables: t.Mapping[str, t.Any] = None):

        self.globals = globals if globals is not None else {}
        self.__initials = initials if initials is not None else {}
        self.__defaults = defaults if defaults is not None else {}
        self.__variables = variables if variables is not None else {}
        self.__cm = ChainMap(self.__variables, self.__defaults, self.__initials)

    def create_new_ctx(self, defaults, initials=None):
        return self.__class__(globals=dict(self.globals), initials=dict(initials or self.__initials), defaults=defaults,
                              variables=self.__variables)

    def find_recursive(self, item):
        value = self.__cm[item]
        try:
            match = self._exp.search(value)
        except:
            return value
        if match:
            item = match.groups()[0]
            return self.find_recursive(item)
        else:
            return value

    def __iter__(self):
        for k, v in self.__cm.items():
            match = None
            try:
                match = self._exp.search(v)
            except:
                yield k, v
            if match:
                yield k, self.find_recursive(match.groups()[0])
            else:
                yield k, v

    def __len__(self):
        return len(self.__cm)

    def __contains__(self, item):
        return item in self.__cm

    def get(self, k, d=None):
        try:
            return self.find_recursive(k)
        except KeyError:
            return d

    def set(self, k, v):
        if k in self.__reserved_keys:
            raise KeyError(f"'{k}' is a reserved key")
        else:
            self.__variables.update({k: v})

    def extract_variables(self):
        return dict(self.__variables)

    def update_variables(self, variables):
        self.__variables.update(variables)


import functools
import re
import sys
import time
import traceback
import typing as t
from contextlib import contextmanager
from http.server import HTTPServer
from io import StringIO
from threading import Thread
from unittest import TestCase, mock
from unittest.mock import Mock

import flask
import requests
from aioresponses import aioresponses, CallbackResult
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from flask.testing import FlaskClient
from flask_jwt_extended import create_access_token

from dm import defaults
from dm.domain.entities import User, update_datemark
from dm.network.auth import HTTPBearerAuth
from dm.utils.helpers import get_entities, get_distributed_entities
from dm.web import create_app, db, errors


def start_mock_server(port, mock_server_request_handler):
    mock_server = HTTPServer(('localhost', port), mock_server_request_handler)
    mock_server_thread = Thread(target=mock_server.serve_forever)
    mock_server_thread.setDaemon(True)
    mock_server_thread.start()


def set_response_from_mock(mock: Mock, url: str, status: int, json: t.Union[str, t.Mapping[str, t.Any]],
                           headers: t.Mapping[str, str] = None):
    """
    Function to mock a requests.HTTP_METHOD
    Parameters
    ----------
    mock
    url
    status
    json
    headers

    Returns
    -------
    None
    """
    resp = requests.Response()
    resp.url = url
    resp.headers = headers or {'USER-AGENT': 'werkzeug/0.16.0', 'CONTENT-TYPE': 'application/json'}
    resp.status_code = status
    resp._content = str(json).encode()
    mock.return_value = resp


def wait_mock_called(mock: Mock, call_count: int, timeout: int = 10):
    """Waits for a mock to be called at least call_count times. Raise a TimeoutError if timeout reached"""
    start = time.time()
    while time.time() < (timeout + start):
        if mock.call_count >= call_count:
            return
        else:
            time.sleep(0.01)
    raise TimeoutError('Timeout reached while waiting for mock to be called')


@contextmanager
def captured_output() -> t.Tuple[StringIO, StringIO]:
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err




def authorization_header(identity='test'):
    access_token = create_access_token(identity=identity)
    return {"Authorization": f"Bearer {access_token}"}


class TestCaseLockBypass(TestCase):

    def run(self, result=None):
        with mock.patch('dm.use_cases.lock.lock'):
            with mock.patch('dm.use_cases.lock.unlock'):
                super().run(result)


def set_callbacks(target: t.List[t.Tuple[str, FlaskClient]], m: aioresponses = None):
    import responses
    method_list = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']

    def requests_callback_client(client, request):
        method_func = getattr(client, request.method.lower())
        try:
            resp = method_func(request.path_url, data=request.body, headers=dict(request.headers))
        except Exception as e:
            return 500, {}, traceback.format_exc()

        return resp.status_code, resp.headers, resp.data

    def callback_client(method, client, url, **kwargs):
        kwargs.pop('allow_redirects')
        # passing headers as a workarround for https://github.com/pnuckowski/aioresponses/issues/111
        func = getattr(client, method.lower())
        try:
            r = func(url.path, headers=kwargs['headers'], json=kwargs.get('json'))
        except Exception as e:
            return CallbackResult(method.upper(), status=500, body=traceback.format_exc(), headers={})

        return CallbackResult(method.upper(), status=r.status_code, body=r.data, content_type=r.content_type,
                              headers=r.headers)

    for dest_regexp, client in target:

        for method in method_list:
            responses.add_callback(method, re.compile(f'https?://{dest_regexp}.*'),
                                   callback=functools.partial(requests_callback_client, client))
        if m:
            for method in method_list:
                func = getattr(m, method.lower())
                func(re.compile(f'https?://{dest_regexp}.*'),
                     callback=functools.partial(callback_client, method, client), repeat=True)


class ValidateResponseMixin:

    def validate_error_response(self, resp: flask.Response, error: errors.BaseError):
        self.assertEqual(error.status_code, resp.status_code)
        self.assertDictEqual(errors.format_error_content(error), resp.get_json())


class TestDimensigonBase(TestCase, ValidateResponseMixin):

    def setUp(self) -> None:
        self.maxDiff = None
        self.app = create_app('test')
        self.app.config['SECURIZER'] = False
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()
        User.set_initial()
        self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()


def generate_dimension_json_data():
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=1024,
        backend=default_backend()
    )
    priv_pem = private_key.private_bytes(encoding=serialization.Encoding.PEM,
                                         format=serialization.PrivateFormat.TraditionalOpenSSL,
                                         encryption_algorithm=serialization.NoEncryption())
    pub_pem = private_key.public_key().public_bytes(encoding=serialization.Encoding.PEM,
                                                    format=serialization.PublicFormat.PKCS1)

    return {'id': '0000000d-0000-0000-0000-000000000001', 'name': 'dimension',
            'private': priv_pem.decode('ascii'),
            'public': pub_pem.decode('ascii'),
            'created_at': defaults.INITIAL_DATEMARK.strftime(defaults.DATETIME_FORMAT), 'current': True}


def load_data(catalog: t.Dict[str, t.List[t.Dict]]):
    e = dict(get_entities())
    de = dict(get_distributed_entities())

    for name, dtos in catalog:
        cls = e[name]
        for dto in dtos:
            if 'last_modified_at' in dto and name in de:
                update_datemark(False)
            o = cls.from_json(dto)
            db.session.add(o)
            db.session.commit()
            if 'last_modified_at' in dto and name in de:
                update_datemark(True)



import time
from unittest import TestCase, mock
from unittest.mock import patch

from flask import url_for
from flask_jwt_extended import create_access_token

from dm import defaults
from dm.domain.entities import Catalog
from dm.domain.entities.bootstrap import set_initial
from dm.domain.entities.locker import State, Locker, Scope
from dm.web import create_app, db


class TestLocker(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        db.create_all()
        set_initial()
        db.session.commit()
        self.datemark = Catalog.max_catalog(str)

        import dm.web.api_1_0.urls.locker as locker_mod
        self.revert_preventing = locker_mod.revert_preventing

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @patch('dm.web.api_1_0.urls.locker.defaults.TIMEOUT_PREVENTING_LOCK', 0.01)
    def test_lock_prevent_timer(self):

        resp = self.client.post(url_for('api_1_0.locker_prevent'),
                                json=dict(scope=Scope.ORCHESTRATION.name, datemark=self.datemark,
                                          applicant=['applicant']),
                                headers=self.headers)

        self.assertEqual(200, resp.status_code)

        l = Locker.query.get(Scope.ORCHESTRATION)
        self.assertEqual(l.state, State.PREVENTING)

        start = time.time()
        while time.time() - start < 5:
            db.session.expire(l)
            if l.state == State.UNLOCKED:
                start -= 5
            else:
                time.sleep(0.01)

        self.assertEqual(State.UNLOCKED, l.state)

    @mock.patch('dm.web.api_1_0.urls.locker.threading')
    def test_lock(self, mock_thread):

        resp = self.client.post(url_for('api_1_0.locker_prevent'),
                                json=dict(scope=Scope.CATALOG.name, datemark=self.datemark,
                                          applicant=['applicant']),
                                headers=self.headers)

        resp = self.client.post(url_for('api_1_0.locker_prevent'),
                                json=dict(scope=Scope.CATALOG.name, datemark=self.datemark,
                                          applicant=['applicant']),
                                headers=self.headers)

        self.assertEqual(409, resp.status_code)

        resp = self.client.post(url_for('api_1_0.locker_prevent'),
                                json=dict(scope=Scope.CATALOG.name, datemark=self.datemark,
                                          applicant='applicant2'),
                                headers=self.headers)

        self.assertEqual(409, resp.status_code)

        # LOWER PRIORITY LOCKER
        resp = self.client.post(url_for('api_1_0.locker_prevent'),
                                json=dict(scope=Scope.UPGRADE.name, datemark=self.datemark, applicant='applicant2'),
                                headers=self.headers)

        self.assertEqual(409, resp.status_code)

        l = Locker.query.get(Scope.CATALOG)
        l.applicant = None
        l.state = State.UNLOCKED

        l = Locker.query.get(Scope.ORCHESTRATION)
        l.applicant = ['applicant']
        l.state = State.PREVENTING
        db.session.commit()

        # HIGHER PRIORITY LOCKER
        resp = self.client.post(url_for('api_1_0.locker_prevent'),
                                json=dict(scope=Scope.CATALOG.name, datemark=self.datemark, applicant='applicant2'),
                                headers=self.headers)

        self.assertEqual(200, resp.status_code)
        l = Locker.query.get(Scope.CATALOG)
        self.assertEqual(l.state, State.PREVENTING)

        # LOCK
        resp = self.client.post(url_for('api_1_0.locker_lock'),
                                json=dict(scope=Scope.ORCHESTRATION.name, applicant='applicant2'),
                                headers=self.headers)

        self.assertEqual(409, resp.status_code)

        resp = self.client.post(url_for('api_1_0.locker_lock'),
                                json=dict(scope=Scope.ORCHESTRATION.name, applicant=['applicant']),
                                headers=self.headers)

        self.assertEqual(200, resp.status_code)
        l = Locker.query.get(Scope.ORCHESTRATION)
        self.assertEqual(l.state, State.LOCKED)

        # LOWER PRIORITY LOCKER
        resp = self.client.post(url_for('api_1_0.locker_lock'),
                                json=dict(scope=Scope.UPGRADE.name, applicant='applicant2'),
                                headers=self.headers)

        self.assertEqual(409, resp.status_code)

        # HIGHER PRIORITY LOCKER
        resp = self.client.post(url_for('api_1_0.locker_lock'),
                                json=dict(scope=Scope.CATALOG.name, applicant='applicant2'),
                                headers=self.headers)

        self.assertEqual(200, resp.status_code)
        l = Locker.query.get(Scope.CATALOG)
        self.assertEqual(l.state, State.LOCKED)

        # UNLOCK
        resp = self.client.post(url_for('api_1_0.locker_unlock'),
                                json=dict(scope=Scope.ORCHESTRATION.name, applicant='applicant2'),
                                headers=self.headers)

        self.assertEqual(409, resp.status_code)

        resp = self.client.post(url_for('api_1_0.locker_unlock'),
                                json=dict(scope=Scope.ORCHESTRATION.name, applicant=['applicant']),
                                headers=self.headers)

        self.assertEqual(200, resp.status_code)
        l = Locker.query.get(Scope.ORCHESTRATION)
        self.assertEqual(l.state, State.UNLOCKED)

    def test_lock_datemark(self):
        resp = self.client.post(url_for('api_1_0.locker_prevent'),
                                json=dict(scope=Scope.ORCHESTRATION.name,
                                          datemark=defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT),
                                          applicant=['applicant']),
                                headers=self.headers)

        self.assertEqual(409, resp.status_code)

    def test_lock_orchestration(self):
        resp = self.client.post(url_for('api_1_0.locker_unlock'),
                                json=dict(scope=Scope.ORCHESTRATION.name, applicant="1"),
                                headers=self.headers)
        self.assertEqual(210, resp.status_code)

        l = Locker.query.get(Scope.ORCHESTRATION)
        l.state = State.PREVENTING
        l.applicant = "1"
        db.session.commit()
        resp = self.client.post(url_for('api_1_0.locker_prevent'),
                                json=dict(scope=Scope.ORCHESTRATION.name, datemark=self.datemark, applicant="1"),
                                headers=self.headers)
        self.assertEqual(210, resp.status_code)

        l.state = State.LOCKED
        db.session.commit()

        resp = self.client.post(url_for('api_1_0.locker_prevent'),
                                json=dict(scope=Scope.ORCHESTRATION.name, datemark=self.datemark, applicant="1"),
                                headers=self.headers)
        self.assertEqual(210, resp.status_code)

        resp = self.client.post(url_for('api_1_0.locker_lock'),
                                json=dict(scope=Scope.ORCHESTRATION.name, applicant="1"),
                                headers=self.headers)
        self.assertEqual(210, resp.status_code)



import copy
import functools
import inspect
import os
import re
import sys
import time
import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime

import flask
import jinja2
import requests
from flask import json
from flask_jwt_extended import create_access_token, get_jwt_identity

from dm import defaults
from dm.domain.entities import Step, Server, Software, OrchExecution
from dm.network.auth import HTTPBearerAuth
from dm.utils import subprocess
from dm.utils.helpers import get_now
from dm.utils.typos import Kwargs
from dm.utils.var_context import VarContext
from dm.web import db, errors
from dm.web.helpers import normalize_hosts
from dm.web.network import request, get


@dataclass
class CompletedProcess:
    success: bool = None
    stdout: t.Union[str, bytes] = None
    stderr: t.Union[str, bytes] = None
    rc: int = None
    start_time: datetime = None
    end_time: datetime = None
    pre_post_error: Exception = None

    def set_start_time(self) -> 'CompletedProcess':
        self.start_time = get_now()
        return self

    def set_end_time(self) -> 'CompletedProcess':
        self.end_time = get_now()
        return self

    def to_json(self):
        return self.__dict__


class IOperationEncapsulation(ABC):

    def __init__(self, code: str, expected_stdout: str = None, expected_stderr: str = None, expected_rc: int = None,
                 system_kwargs: Kwargs = None):
        """
        Operation Initializer

        Parameters
        ----------
        code:
            Code to be executed
        expected_output:
            Expected output to compare from execution output. If None it won't be compared
        expected_rc:
            Expected return code to be compared with the execution return code. If None it won't be compared
        system_kwargs:
            system arguments to be passed to the commend execution, not the variables inside the code
        """
        self.code = code
        self.expected_stdout = expected_stdout
        self.expected_stderr = expected_stderr
        self.expected_rc = expected_rc
        self.system_kwargs = system_kwargs or {}

    def load_code(self):
        return

    @abstractmethod
    def execute(self, var_context: VarContext, timeout=None) -> CompletedProcess:
        """
        StepExecution process
        Parameters
        ----------
        var_context:
            var_context to be passed through the execution

        Returns
        -------
        StepExecution:
            dataclass containing all the information from the result execution
        """

    def rpl_params(self, params: Kwargs):
        template = jinja2.Template(self.code)
        return template.render(params)

    def evaluate_result(self, cp: CompletedProcess):
        if cp.success is None:
            res = []
            if self.expected_stdout is not None:
                if isinstance(cp.stdout, str):
                    res.append(True) if re.search(self.expected_stdout, cp.stdout) else res.append(False)
            if self.expected_stderr is not None:
                if isinstance(cp.stderr, str):
                    res.append(True) if re.search(self.expected_stderr, cp.stderr) else res.append(False)
            if self.expected_rc is not None:
                res.append(True) if self.expected_rc == cp.rc else res.append(False)
            cp.success = all(res)
        return cp


class AnsibleOperation(IOperationEncapsulation):

    def execute(self, var_context: VarContext, timeout=None) -> CompletedProcess:
        code = self.code
        if re.match(r'^[^<>:;,?"*|/\\]+$', self.code):
            file = os.path.join('', 'ansible', self.code)
            if os.path.exists(file):
                with open(file, 'r') as fh:
                    code = fh.read()

        template = self.rpl_params(dict(var_context))

        system_kwargs = self.system_kwargs.copy()

        tokens = ('ansible-playbook', '-i', '"localhost,"', '-c', 'local') + tuple(template)

        cp = CompletedProcess()
        cp.set_start_time()
        try:
            r = subprocess.run(tokens, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               **system_kwargs, timeout=timeout)
            cp.stdout = r.stdout
            cp.stderr = r.stderr
            cp.rc = r.returncode
        except (subprocess.TimeoutExpired, ValueError) as e:
            cp.stderr = f"{e.__class__.__name__}{e.args}"
            cp.success = False
        finally:
            cp.set_end_time()

        return self.evaluate_result(cp)


class RequestOperation(IOperationEncapsulation):

    def execute(self, var_context: VarContext, timeout=None) -> CompletedProcess:
        params_dict = dict(var_context)
        kwargs = json.loads(self.rpl_params(params_dict))
        kwargs.update(self.system_kwargs)
        cp = CompletedProcess()
        cp.set_start_time()

        # common parameters
        if kwargs.get('timeout') is not None and timeout is not None:
            kwargs['timeout'] = min(kwargs.get('timeout'), timeout)
        method = kwargs.pop('method').lower()

        resp, exception = None, None

        if 'software_id' in var_context:
            def search_cost(ssa, route_list):
                cost = [route['cost'] for route in route_list if str(ssa.server.id) == route['destination_id']]
                if cost:
                    if cost[0] is None:
                        cost = 999999
                    else:
                        cost = cost[0]
                else:
                    cost = 999999
                return cost

            soft = Software.query.get(params_dict.get('software_id', None))
            if not soft:
                cp.stderr = f"software id '{params_dict.get('software_id', None)}' not found"
                cp.success = False
                cp.set_end_time()
                return cp
            if not soft.ssas:
                cp.stderr = f"{soft.id} has no server association"
                cp.success = False
                cp.set_end_time()
                return cp
            dest_server = Server.query.get(params_dict.get('server_id', None))
            if not dest_server:
                cp.stderr = f"destination server id '{params_dict.get('server_id', None)}' not found"
                cp.success = False
                cp.set_end_time()
                return cp

            auth = HTTPBearerAuth(create_access_token(get_jwt_identity()))
            # decide best server source
            resp = get(dest_server, 'api_1_0.routes', auth=auth, timeout=5)
            if resp.code == 200:
                ssas = copy.copy(soft.ssas)
                ssas.sort(key=functools.partial(search_cost, route_list=resp.msg['route_list']))
            else:
                ssas = soft.ssas
            server = ssas[0].server

            # Process kwargs
            if 'auth' not in kwargs:
                kwargs['auth'] = auth

            view_or_url = kwargs.pop('view_or_url', kwargs.pop('view', kwargs.pop('url', None)))

            # run request
            resp = request(method, server, view_or_url, **kwargs)
            cp.stdout = flask.json.dumps(resp.msg) if isinstance(resp.msg, dict) else resp.msg
            cp.stderr = str(resp.exception) if str(resp.exception) else resp.exception.__class__.__name__
            cp.rc = resp.code
            if resp.exception is None:
                self.evaluate_result(cp)
        else:
            # process kwargs
            url = kwargs.pop('view_or_url', kwargs.pop('url', None))

            # Run request
            try:
                resp = requests.request(method, url, **kwargs)
            except Exception as e:
                exception = e
            else:
                cp.rc = resp.status_code
                cp.stdout = resp.text
            if exception is None:
                self.evaluate_result(cp)
            else:
                cp.success = False
                cp.stderr = str(exception) if str(exception) else exception.__class__.__name__
        cp.set_end_time()
        return cp

class NativeOperation(IOperationEncapsulation):

    def execute(self, var_context: VarContext, timeout=None):
        pass


class NativeWaitOperation(IOperationEncapsulation):

    def execute(self, var_context: VarContext, timeout=None):
        start = time.time()
        found = []
        cp = CompletedProcess()
        cp.set_start_time()
        try:
            min_timeout = min(timeout, var_context.get('timeout', None))
        except TypeError:
            if timeout is not None:
                min_timeout = timeout
            elif var_context.get('timeout', None) is not None:
                min_timeout = var_context.get('timeout')
            else:
                min_timeout = defaults.MAX_TIME_WAITING_SERVERS
        for sn in var_context.get('list_server_names', []):
            count = db.session.query(Server).filter_by(name=sn).count()
            if count == 1:
                found.append(sn)
                continue
            while (time.time() - start) < min_timeout:
                count = db.session.query(Server).filter_by(name=sn).count()
                if count == 0:
                    time.sleep(self.system_kwargs.get('sleep_time', 2))
                else:
                    found.append(sn)
                    break

        if var_context.get('list_server_names', []) == found:
            cp.success = True
            cp.stdout = f"Servers {', '.join(var_context.get('list_server_names', []))} found"
        else:
            cp.success = False
            not_found = set(var_context['list_server_names']) - set(found)
            cp.stderr = f"Servers {', '.join(not_found)} not created after {min_timeout} seconds"
        cp.set_end_time()
        return cp


class PythonOperation(IOperationEncapsulation):
    def execute(self, var_context: VarContext, timeout=None) -> CompletedProcess:
        pass


class ShellOperation(IOperationEncapsulation):

    def execute(self, var_context: VarContext, timeout=None) -> CompletedProcess:
        tokens = self.rpl_params(dict(var_context))

        system_kwargs = self.system_kwargs.copy()

        timeout = system_kwargs.pop('timeout', 300)

        cp = CompletedProcess()
        cp.set_start_time()
        r = None
        try:
            r = subprocess.run(tokens, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               **system_kwargs, timeout=timeout)
            cp.stdout = r.stdout.decode() if isinstance(r.stdout, bytes) else r.stdout
            cp.stderr = r.stderr.decode() if isinstance(r.stderr, bytes) else r.stderr
            cp.rc = r.returncode
        except (subprocess.TimeoutExpired, ValueError) as e:
            cp.stderr = f"{e.__class__.__name__}{e.args}"
            cp.success = False
        finally:
            cp.set_end_time()

        self.evaluate_result(cp)

        return cp


class OrchestrationOperation(IOperationEncapsulation):

    def execute(self, var_context: VarContext, timeout=None) -> CompletedProcess:
        from dm.web.async_functions import deploy_orchestration

        exe = OrchExecution(orchestration_id=var_context.get('orchestration_id'),
                            target=var_context.get('hosts'),
                            params=dict(var_context),
                            parent_orch_execution_id=var_context.globals.get('execution_id'),
                            executor_id=var_context.globals.get('executor_id'),
                            server=Server.get_current())
        db.session.add(exe)
        db.session.commit()
        cp = CompletedProcess().set_start_time()
        hosts = copy.deepcopy(var_context.get('hosts'))
        not_found = normalize_hosts(hosts)
        if not_found:
            cp.stderr = str(errors.ServerNormalizationError(not_found))
            cp.success = False
        else:
            try:
                deploy_orchestration(orchestration=var_context.get('orchestration_id'), hosts=hosts,
                                     var_context=var_context.create_new_ctx({}, initials=dict(var_context)),
                                     execution=exe)
            except Exception as e:
                cp.stderr = str(e) if str(e) else e.__class__.__name__
                cp.success = False
            else:
                db.session.refresh(exe)
                cp.success = exe.success

        return cp.set_end_time()


from dm.domain.entities import ActionType

_operation_classes = {}
for name, cls in inspect.getmembers(sys.modules['dm.use_cases.operations'],
                                    lambda x: (inspect.isclass(x) and issubclass(x, IOperationEncapsulation))):
    _operation_classes.update({name: cls})

_factories: t.Dict['ActionType', t.Type[IOperationEncapsulation]] = {}

for at in ActionType:
    try:
        _factories.update({at: _operation_classes[at.name.capitalize() + 'Operation']})
    except KeyError:
        NotImplementedError(f"{at.name.capitalize() + 'Operation'} not implemented")


def create_operation(step: 'Step') -> IOperationEncapsulation:
    kls = _factories[step.action_type]

    if kls == NativeOperation:
        if step.action_template.name == 'wait':
            kls = NativeWaitOperation
    return kls(code=step.code, expected_stdout=step.expected_stdout, expected_stderr=step.expected_stderr,
               expected_rc=step.expected_rc,
               system_kwargs=step.system_kwargs)



import os
import typing as t

from prompt_toolkit.history import History

from dm.utils.helpers import get_now


class FileTagHistory(History):
    """
    :class:`.History` class that stores all strings in a file.
    """

    def __init__(self, filename: str, tag: str) -> None:
        self.tag = tag
        self.filename = filename
        super(FileTagHistory, self).__init__()

    def load_history_strings(self) -> t.Iterable[str]:
        strings: t.List[str] = []
        lines: t.List[str] = []

        def add() -> None:
            if lines:
                # Join and drop trailing newline.
                string = "".join(lines)[:-1]

                strings.append(string)

        if os.path.exists(self.filename):
            with open(self.filename, "rb") as f:
                for line_bytes in f:
                    line = line_bytes.decode("utf-8")

                    if line.startswith(f"{self.tag}+"):
                        lines.append(line.split('+', 1)[1])
                    else:
                        add()
                        lines = []

                add()

        # Reverse the order, because newest items have to go first.
        return reversed(strings)

    def store_string(self, string: str) -> None:
        # Save to file.
        with open(self.filename, "ab") as f:
            def write(t: str) -> None:
                f.write(t.encode("utf-8"))

            write("\n# %s\n" % get_now())
            for line in string.split("\n"):
                write("%s+%s\n" % (self.tag, line))
